# -*- coding: utf-8 -*-
u"""Default config

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function

def alpha():
    return {}


def beta():
    return {}


def dev():
    return {
        'p1': {
            'm1': {
                'p6': '2012-12-12T12:12:12Z',
            },
        },
    }


def prod():
    return {}



"""test http server

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import pytest


@pytest.mark.asyncio
async def test_basic():
    from pykern.api import unit_util

    async with unit_util.Setup(api_classes=(_class(),)) as c:
        from pykern.pkcollections import PKDict
        from pykern import pkunit, pkdebug

        e = PKDict(ping="pong")
        pkunit.pkeq(e.pkupdate(counter=1), await c.call_api("echo", e))
        pkunit.pkeq(e.pkupdate(counter=2), await c.call_api("echo", e))


@pytest.mark.asyncio
async def test_subscribe():
    from pykern.api import unit_util

    async def _client_unsub(iter_count, sub):
        sub.unsubscribe()
        with pkunit.pkexcept(util.APIDisconnected):
            pkunit.pkeq(None, await sub.result_get())
        sub.unsubscribe()

    async def _no_values(iter_count, sub):
        pkunit.pkeq(None, await sub.result_get())
        sub.unsubscribe()

    async def _server_unsub(iter_count, sub):
        for i in range(iter_count):
            pkunit.pkeq(PKDict(count=i), await sub.result_get())
        pkunit.pkeq(None, await sub.result_get())
        with pkunit.pkexcept(util.APIDisconnected):
            pkunit.pkeq(None, await sub.result_get())

    async with unit_util.Setup(api_classes=(_class(),)) as c:
        from pykern.pkcollections import PKDict
        from pykern import pkunit
        from pykern.api import util
        import asyncio

        for n, o in (
            (3, _server_unsub),
            (1, _client_unsub),
            (0, _no_values),
        ):
            with await c.subscribe_api("sub1", PKDict(iter_count=n)) as s:
                await o(n, s)


def _class():
    from pykern.api import util
    from pykern.pkcollections import PKDict
    from pykern import quest
    import asyncio

    class _API(quest.API):

        async def api_echo(self, api_args):
            self.session.pksetdefault(counter=0).counter += 1
            return api_args.pkupdate(counter=self.session.counter)

        @util.subscription
        async def api_sub1(self, api_args):
            await asyncio.sleep(0.1)
            for i in range(api_args.iter_count):
                await asyncio.sleep(0.2)
                if self.is_quest_end():
                    break
                self.subscription.result_put(PKDict(count=i))
            return None

    return _API



"""?

:copyright: Copyright (c) 2025 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.api import util
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp
import pykern.quest


class AuthAPI(pykern.quest.API):

    #: Defaults version number but allows override
    VERSION = util.AUTH_API_VERSION

    async def api_authenticate_connection(self, api_args):
        """Process AuthRequest from server

        api_args:
            token (str): secret value evaluated by `AuthAPI`
            version (int): protocol version
        Args:
            api_args (PKDict): what to validate
        Returns:
            Result: validation result
        """
        if (v := api_args.get("version")) != self.VERSION:
            raise util.APIProtocolError(f"invalid version={v}, expected={self.VERSION}")
        if (t := self.token()) is not None and t != api_args.token:
            # Do not log token
            raise util.APIForbidden()
        return PKDict()

    def token(self):
        return None



# -*- coding: utf-8 -*-
"""Front-end command line for :mod:`xyzzy1`.

See :mod:`pykern.pkcli` for how this module is used.

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pykern.pkcli
import sys


def main():
    return pykern.pkcli.main("xyzzy1")


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
"""xyzzy1 setup script

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pykern.pksetup

pykern.pksetup.setup(
    name="xyzzy1",
    author="RadiaSoft LLC",
    author_email="pip@radiasoft.net",
    description="a magic trick in adventure",
    install_requires=[
        "pykern",
    ],
    license="http://www.apache.org/licenses/LICENSE-2.0.html",
    url="https://github.com/radiasoft/xyzzy1",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Topic :: Utilities",
    ],
)



"""Wrapper for :mod:`yaml`

:copyright: Copyright (c) 2015-2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdp, pkdlog
from pykern import pkcompat
from pykern import pkinspect
from pykern import pkio
from pykern import pkresource
import ruamel.yaml


#: file extension for yaml
PATH_EXT = ".yml"


def dump_pretty(obj, filename, pretty=True, **kwargs):
    """Formats as yaml as string

    If an object is not encoded by default, will call str() on the
    object.

    Unlike `pkjson.dump_pretty`, returns nothing.

    Args:
        obj (object): any Python object
        filename (str or py.path): where to write
        pretty (bool): pretty print [True]
        kwargs (object): other arguments to `ruamel.yaml.dump`
    """
    try:
        y = ruamel.yaml.YAML()
        if pretty:
            y.indent(mapping=2, sequence=4, offset=2)
            y.dump(
                _fixup_dump(obj), stream=pkio.open_text(filename, mode="wt"), **kwargs
            )
    except Exception:
        pkdlog("error writing file={}", filename)
        raise


def load_file(filename):
    """Read a file, making sure all keys and values are locale.

    Args:
        filename (str or py.path): file to read (Note: ``.yml`` will not be appended)

    Returns:
        object: `PKDict` or list
    """
    try:
        return load_str(pkio.read_text(filename))
    except Exception:
        pkdlog("error file={}", filename)
        raise


def load_resource(basename):
    """Read a resource, making sure all keys and values are locale

    Args:
        basename (str): file to read without yml suffix

    Returns:
        object: `PKDict` or list
    """
    return load_file(
        pkresource.filename(basename + PATH_EXT, pkinspect.caller_module()),
    )


def load_str(value):
    """Read a value, making sure all keys and values are locale.

    Args:
        value (str): string to parse

    Returns:
        object: `PKDict` or list
    """
    return _fixup_load(
        ruamel.yaml.YAML(typ="safe").load(value),
    )


def _fixup_dump(obj):
    if isinstance(obj, PKDict):
        return {k: _fixup_dump(v) for k, v in obj.items()}
    if isinstance(obj, (list, tuple)):
        return [_fixup_dump(v) for v in obj]
    return obj


def _fixup_load(obj):
    """Convert all objects to locale strings and PKDict"""

    def _scalar(v):
        if isinstance(v, (bytes, bytearray)):
            return pkcompat.locale_str(v)
        return v

    if isinstance(obj, dict):
        return PKDict({_scalar(k): _fixup_load(v) for k, v in obj.items()})
    if isinstance(obj, (list, tuple)):
        return [_fixup_load(v) for v in obj]
    return _scalar(obj)



# -*- coding: utf-8 -*-
"""pytest for `pykern.pkcli.projex`

:copyright: Copyright (c) 2015-2023 RadiaSoft, LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import py
import pytest
import re
import subprocess


def test_init_rs_tree():
    """Normal case"""
    from pykern import pkio
    from pykern.pkcli import projex
    from pykern import pkunit

    with pkunit.save_chdir_work():
        name = "rs_proj1"
        pkio.mkdir_parent(name)
        with pkio.save_chdir(name):
            subprocess.check_call(["git", "init", "."])
            subprocess.check_call(["git", "config", "user.email", "pip@pykern.org"])
            subprocess.check_call(["git", "config", "user.name", "pykern"])
            projex.init_rs_tree(
                description="some radiasoftee project",
            )
            for expect_fn, expect_re in (
                ("LICENSE", "Apache License"),
                ("pyproject.toml", 'name = "rs_proj1"'),
            ):
                assert re.search(
                    expect_re, pkio.read_text(expect_fn)
                ), '{} should exist and match "{}"'.format(expect_fn, expect_re)


def test_init_tree():
    """Normal case"""
    from pykern import pkio
    from pykern.pkcli import projex
    from pykern import pkunit

    with pkunit.save_chdir_work():
        name = "proj1"
        pkio.mkdir_parent(name)
        with pkio.save_chdir(name):
            subprocess.check_call(["git", "init", "."])
            subprocess.check_call(["git", "config", "user.email", "pip@pykern.org"])
            subprocess.check_call(["git", "config", "user.name", "pykern"])
            projex.init_tree(
                name=name,
                author="zauthor",
                author_email="author@example.com",
                description="some python project",
                license="MIT",
                url="http://example.com",
            )
            pkio.write_text("tests/test_1.py", "def test_1(): pass")
            for expect_fn, expect_re in (
                (".gitignore", "MANIFEST.in"),
                ("LICENSE", "The MIT License"),
                ("README.md", "licenses/MIT"),
                ("docs/_static/.gitignore", ""),
                ("docs/_templates/.gitignore", ""),
                ("docs/index.rst", name),
                ("pyproject.toml", 'name = "zauthor"'),
                ("proj1/proj1_console.py", r":copyright:.*zauthor\."),
                ("tests/.gitignore", "_work"),
                (name + "/__init__.py", ""),
                (name + "/package_data/.gitignore", ""),
                (
                    "{}/{}_console.py".format(name, name),
                    r'main\("{}"\)'.format(name),
                ),
            ):
                assert re.search(
                    expect_re, pkio.read_text(expect_fn)
                ), '{} should exist and match "{}"'.format(expect_fn, expect_re)



"""Generate documentation with Sphinx

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp
import email.utils
import importlib.metadata
import pykern.pkcli
import pykern.pkcollections
import pykern.pkio
import pykern.pkresource
import pykern.pksubprocess
import re
import toml

_DOCS_SUBDIR = "docs"


def prepare(package):
    """Create conf.py and run ``sphinx-apidoc``

    conf.py is generated from `package` metadata so `package` must be
    installed.

    Args:
        package (str): package to build

    """

    def _conf(substitutions):
        rv = pykern.pkio.read_text(pykern.pkresource.file_path("sphinx-conf.py"))
        for k, v in substitutions.items():
            rv = rv.replace("$" + k, v)
        return rv

    def _parse_metadata():
        try:
            return _vars(importlib.metadata.metadata(package))
        except importlib.metadata.PackageNotFoundError as e:
            pykern.pkcli.command_error(
                f"unable to import package={package}; run pip install ."
            )

    def _vars(metadata):
        return _year(
            PKDict(
                author=email.utils.parseaddr(metadata["Author-email"])[0],
                name=metadata["Name"],
                version=metadata["Version"],
            ),
        )

    def _year(rv):
        m = re.search(r"^(\d{4})\d{4}\.\d+$", rv.version)
        if m:
            return rv.pkupdate(year=m.group(1))
        pykern.pkcli.command_error(
            f"unable to parse version={rv.version} package={rv.name}"
        )

    d = pykern.pkio.py_path(_DOCS_SUBDIR)
    if not d.check(dir=True):
        pykern.pkcli.command_error(
            f"missing directory={d.basename}; run from Python root directory"
        )
    v = _parse_metadata()
    pykern.pkio.write_text(d.join("conf.py"), _conf(v))
    pykern.pksubprocess.check_call_with_signals(
        ("sphinx-apidoc", "--force", "--module-first", "-o", d.basename, v.name),
    )


def build(builder="html"):
    """`prepare` and run ``sphinx-build`` for html

    Reads ``pyproject.toml`` to get project name. Creates ``_build_<builder>`` subdirectory
    with the output.

    Args:
        builder (str): types of builders. See `Sphinx builder list <https://www.sphinx-doc.org/en/master/usage/builders/index.html>`_ [html]

    """

    try:
        t = toml.load("pyproject.toml")
    except Exception as e:
        if pykern.pkio.exception_is_not_found(e):
            pykern.pkcli.command_error(
                "pyproject.toml not found; convert setup.py (if exists) or run from Python root directory"
            )
        raise
    prepare(t["project"]["name"])
    pykern.pksubprocess.check_call_with_signals(
        (
            "sphinx-build",
            "-M",
            builder,
            _DOCS_SUBDIR,
            f"_build_{builder}",
            # --fail-on-warning not yet in this version
            "-W",
        ),
    )



# -*- coding: utf-8 -*-
"""Force a coroutine never awaited warning

:copyright: Copyright (c) 2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_coroutine_never_awaited():
    import asyncio

    async def coroutine():
        asyncio.sleep(0)

    asyncio.run(coroutine())



from pykern.pkcollections import PKDict


def add_score(self, num):
    return self.fconf_var('score') + num


def four(self):
    return 4


def must_be_pkdict(self, arg):
    assert isinstance(arg, PKDict)
    return "was_pkdict"


def mymap(self, func, *args):
    res = []
    for a in args:
        res.extend(map(func, a))
    return res


def nop(self, arg):
    return arg


def three(self, one):
    return PKDict(one=one, four=self.four())



"""Various constant values

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

import py.path
import re


#: DEPRECATED use `str`
STRING_TYPES = str

#: Class of `pkio.py_path` and `py.path.local`
PY_PATH_LOCAL_TYPE = type(py.path.local())

#: Copied from numconv, which copied from RFC1924
BASE62_CHARS = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"

#: IP address for local servers
LOCALHOST_IP = "127.0.0.1"

#: The subdirectory in the top-level Python where to put resources
PACKAGE_DATA = "package_data"

#: Use this (sparingly, pkdlog is prefered) when you want to use print directly.
builtin_print = print



# -*- coding: utf-8 -*-
"""test fconf

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_parser():
    from pykern import pkunit, fconf, pkio, pkjson
    from pykern.pkcollections import PKDict
    from pykern.pkdebug import pkdp

    for d in pkunit.case_dirs():
        p = fconf.Parser(
            pkio.sorted_glob("*.py") + pkio.sorted_glob("*.yml"),
            base_vars=(
                PKDict(fconf_test=PKDict(var1=1234)) if d.basename == "1" else None
            ),
        )
        r = p.result
        r.macros = [f'{p.name}({",".join(p.params)})' for p in p.macros.values()]
        pkjson.dump_pretty(r, filename="res.json")



# -*- coding: utf-8 -*-
""":mod:`xyzzy1` package

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pkg_resources

try:
    # We only have a version once the package is installed.
    __version__ = pkg_resources.get_distribution("xyzzy1").version
except pkg_resources.DistributionNotFound:
    pass



"""Where external resources are stored

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# Root module: avoid importing modules which import pkconfig
from pykern import pkconst
from pykern import pkinspect
from pykern import pkio
import errno
import glob
import importlib
import pkg_resources
import os.path


def file_path(relative_filename, caller_context=None, packages=None):
    """Return the path to the resource

    Args:
        relative_filename (str): file name relative to package_data directory.
        caller_context (object): Any object from which to get the `root_package`
        packages (List[str]): Packages to search.

    Returns:
        py.path: absolute path of the resource file
    """
    return pkio.py_path(filename(relative_filename, caller_context, packages))


def filename(relative_filename, caller_context=None, packages=None):
    """Return the filename to the resource

    Args:
        relative_filename (str): file name relative to package_data directory.
        caller_context (object): Any object from which to get the `root_package`
        packages (List[str]): Packages to search.

    Returns:
        str: absolute path of the resource file
    """
    assert not os.path.isabs(
        relative_filename
    ), "must not be an absolute file name={}".format(relative_filename)
    a = []
    for f, p in _files(relative_filename, caller_context, packages):
        a.append(p)
        if os.path.exists(f):
            return f
    _raise_no_file_found(a, relative_filename)


def glob_paths(relative_path, caller_context=None, packages=None):
    """Find all paths that match the relative path in all packages

    Args:
        relative_path(str): Path relative to package_data directory.
        caller_context (object): Any object from which to get the `root_package`.
        packages (List[str]): Packages to search.
    Returns:
        py.path: absolute paths of the matched files
    """
    r = []
    a = []
    for f, p in _files(relative_path, caller_context, packages):
        a.append(p)
        r.extend(glob.glob(f))
    return [pkio.py_path(f) for f in r]


def _files(path, caller_context, packages):
    if caller_context and packages:
        raise ValueError(
            f"Use only one of caller_context={caller_context} and packages={packages}",
        )
    for p in list(
        map(
            lambda m: pkinspect.root_package(importlib.import_module(m)),
            packages
            or [
                pkinspect.root_package(
                    caller_context if caller_context else pkinspect.caller_module()
                )
            ],
        )
    ):
        yield (
            # Will be fixed in https://github.com/radiasoft/pykern/issues/462
            pkg_resources.resource_filename(
                p,
                os.path.join(pkconst.PACKAGE_DATA, path),
            ),
            p,
        )


def _raise_no_file_found(packages, path):
    msg = f"unable to locate in packages={packages}"
    if "__main__" in packages:
        msg += "; do not call module as a program"
    raise IOError(errno.ENOENT, msg, path)



# -*- coding: utf-8 -*-
"""private setup script

:copyright: Copyright (c) 2024 James Bond.  All Rights Reserved.
:license: PROPRIETARY AND CONFIDENTIAL. See LICENSE file for details.
"""
import pykern.pksetup

pykern.pksetup.setup(
    name="private",
    author="James Bond",
    author_email="007@example.com",
    description="Q's super private project",
    install_requires=[
        "pykern",
    ],
    entry_points={
        "console_scripts": ["qdev=private.qdev:main"],
    },
    license="PROPRIETARY AND CONFIDENTIAL. See LICENSE file for details.",
    url="https://example.com/jb",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: Other/Proprietary License",
        "Programming Language :: Python",
        "Topic :: Utilities",
    ],
)



"""test that links with the same name have to have same fmt

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkcollections import PKDict
import pykern.xlsx

PATH = "case3.xlsx"
w = pykern.xlsx.Workbook(path=PATH)
s = w.sheet(title="s1")
t = s.table(title="t1")
for i in range(3):
    t.row(
        count=t.cell(i, link="count"),
    )
i += 1
t.row(
    # different fmt than the above
    count=t.cell(i, link="count", fmt="text"),
)
w.save()



# -*- coding: utf-8 -*-
"""pytest to see if extras_require worked.

:copyright: Copyright (c) 2016 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function

def test_import():
    import adhan
    import civicjson



last_self = None
last_cmd = None
last_arg = None


def should_not_find():
    pass


class Commands:
    def __init__(self):
        global last_self

        last_self = self

    def cmd1(self, arg1):
        global last_cmd, last_arg
        last_cmd = self.cmd1.__name__
        last_arg = arg1

    def _should_not_find():
        pass



"""WebSocket Quest client

:copyright: Copyright (c) 2025 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.api import util
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp, pkdexc
import asyncio
import pykern.util
import tornado.httpclient
import tornado.websocket


class Client:
    """Wrapper for `tornado.httpclient.AsyncHTTPClient`

    Maybe called as an async context manager

    `http_config.request_config` is deprecated.

    Args:
        http_config (PKDict): tcp_ip, tcp_port, api_uri
    """

    def __init__(self, http_config):
        # TODO(robnagler) tls with verification(?)
        self.uri = (
            f"ws://{http_config.tcp_ip}:{http_config.tcp_port}{http_config.api_uri}"
        )
        self._authenticated = False
        self._connection = None
        self._destroyed = False
        self._next_call_id = 1
        self._pending_calls = PKDict()

    async def call_api(self, api_name, api_args):
        """Make a request to the API server

        Args:
            api_name (str): what to call on the server
            api_args (PKDict): passed verbatim to the API on the server.
        Returns:
            str: value of `api_result`.
        Raises:
           pykern.util.APIError: if there was an raise in the API or on a server protocol violation
           Exception: other exceptions that `AsyncHTTPClient.fetch` may raise, e.g. NotFound
        """

        return await self._send_api(api_name, api_args, util.MsgKind.CALL).result_get()

    async def connect(self, auth_args=None):
        """Connect to the server

        Args:
            auth_args (PKDict): how to authenticate connection; may be AuthArgs or other PKDict [None]
        Returns:
            AuthReply: status of connection
        """

        async def _auth():
            try:
                await self.call_api(util.AUTH_API_NAME, _auth_args())
                return True
            except Exception as e:
                if self._destroyed:
                    return False
                self.destroy()
                raise

        def _auth_args():
            rv = auth_args
            if rv is None:
                rv = PKDict()
            return rv.pksetdefault(token=None, version=util.AUTH_API_VERSION)

        if self._destroyed:
            raise AssertionError("destroyed")
        if self._connection:
            raise AssertionError("already connected")
        self._connection = await tornado.websocket.websocket_connect(
            tornado.httpclient.HTTPRequest(self.uri, method="GET"),
            # TODO(robnagler) accept in http_config. share defaults with sirepo.job.
            max_message_size=int(2e8),
            ping_interval=120,
            ping_timeout=240,
        )
        asyncio.create_task(self._read_loop())
        self._authenticated = await _auth()

    def destroy(self):
        """Must be called"""
        if self._destroyed:
            return
        # Allow functions to call back so _destroyed is still True.
        # Reversed so we unsubscribe in opposite order of subscribe
        for c in reversed(list(self._pending_calls.values())):
            try:
                c.destroy()
            except Exception as e:
                pkdlog("{} destroy exception={}", c, e)
        # Just in case of exceptions above
        self._pending_calls = None
        self._destroyed = True
        if self._connection:
            self._connection.close()
            self._connection = None

    def remove_call(self, call_id):
        """Not a public interface"""
        if self._destroyed:
            return
        return self._pending_calls.pkdel(call_id)

    async def subscribe_api(self, api_name, api_args):
        """Subscribe to api_name from API server

        Maybe used in ``with``::

            with client.subscribe_api(api, args) as s:
                while (r := await s.result_get()):
                    process r

        Alternately, you must call `_Call.unsubscribe`::

            s = client.subscribe_api(api, args)
            r = await s.result_get()
            ... process r and possibly more calls to result_get ...
            s.unsubscribe()

        Args:
            api_name (str): what to call on the server
            api_args (PKDict): passed verbatim to the API on the server.
        Returns:
            _Call: to get replies or unsubscribe
        """

        return self._send_api(api_name, api_args, util.MsgKind.SUBSCRIBE)

    def unsubscribe_call(self, call_id):
        """Not a public interface"""
        if self._destroyed:
            return
        if self.remove_call(call_id):
            self._send_msg(PKDict(call_id=call_id, msg_kind=util.MsgKind.UNSUBSCRIBE))

    async def __aenter__(self):
        await self.connect()
        if self._destroyed:
            raise util.APIDisconnected()
        return self

    async def __aexit__(self, *args, **kwargs):
        self.destroy()
        return False

    async def _read_loop(self):
        def _unpack(msg):
            if msg is None:
                return None
            r, e = util.msg_unpack(msg, "client")
            if e:
                pkdlog("msg_unpack error={} {}", e, self)
                return None
            return r

        m = r = None
        try:
            if self._destroyed:
                return
            while m := await self._connection.read_message():
                if self._destroyed:
                    return
                if not (r := _unpack(m)):
                    break
                if not (c := self._pending_calls.get(r.call_id)):
                    pkdlog("call_id={} not found {}", r.call_id, self)
                    # May happen on subscriptions
                    continue
                c.reply_put(r)
                if not c.is_subscription or r.msg_kind.is_unsubscribe():
                    # Call is no longer valid for messages
                    self.remove_call(r.call_id)
                # Clear all state before next await
                c = m = r = None
        except Exception as e:
            pkdlog("exception={} reply={} stack={}", e, r, pkdexc())
        try:
            self.destroy()
        except Exception as e:
            pkdlog("exception={} stack={}", e, pkdexc())

    def __repr__(self):
        def _calls():
            if not self._pending_calls:
                return ""
            return " " + " ".join(map(str, self._pending_calls.values()))

        if self._destroyed:
            return f"<{self.__class__.__name__} DESTROYED>"
        return f"<{self.__class__.__name__}{_calls()}>"

    def _send_api(self, api_name, api_args, msg_kind):
        def _send():
            c = PKDict(
                api_name=api_name,
                api_args=api_args,
                call_id=self._next_call_id,
                msg_kind=msg_kind,
            )
            self._next_call_id += 1
            rv = _Call(self, c)
            self._pending_calls[c.call_id] = rv
            self._send_msg(c)
            return rv

        if self._destroyed:
            raise util.APIDisconnected()
        if self._connection is None:
            raise AssertionError("no connection, must call connect() first")
        if not self._authenticated and api_name != util.AUTH_API_NAME:
            raise AssertionError(
                "connection not authenticated; wait for connect() to return"
            )
        return _send()

    def _send_msg(self, msg):
        self._connection.write_message(util.msg_pack(msg), binary=True)


class _Call:
    """Holds state of an API call

    Attributes:
        api_name (str): name of api
        is_subscription (bool): True if call is subscribed
    """

    def __init__(self, client, msg):
        self.api_name = msg.api_name
        self.is_subscription = msg.msg_kind.is_subscribe()
        self._call_id = msg.call_id
        self._client = client
        # TODO(robnagler) should there be a maximum?
        self._reply_q = asyncio.Queue()
        self._destroyed = False

    def destroy(self):
        if self._destroyed:
            return
        self._destroyed = True
        self._client.remove_call(self._call_id)
        if x := getattr(self._reply_q, "shutdown", None):
            x.shutdown(immediate=True)
        else:
            # Inferior to shutdown, but necessary pre-Python 3.13
            self._reply_q.put_nowait(None)
        self._client = None
        self._reply_q = None

    def reply_put(self, msg):
        if self._destroyed:
            return
        self._reply_q.put_nowait(msg)

    async def result_get(self):
        """Get the next result from a subscription.

        Used internally to this module to get a result from a one-time call.

        Returns:
            PKDict|None: If None, subscription ended normally.
        """
        if self._destroyed:
            raise util.APIDisconnected()
        d = True
        try:
            try:
                rv = await self._reply_q.get()
            except Exception as e:
                if (x := getattr(asyncio, "QueueShutDown", None)) and isinstance(e, x):
                    raise util.APIDisconnected()
                raise
            if self._destroyed:
                raise util.APIDisconnected()
            self._reply_q.task_done()
            if rv is None:
                raise util.APIDisconnected()
            if rv.msg_kind.is_unsubscribe():
                return None
            if rv.api_error:
                raise pykern.util.APIError(rv.api_error)
            d = not self.is_subscription
            return rv.api_result
        finally:
            if d:
                # all cases above are implicit unsubscribe
                self.destroy()

    def unsubscribe(self):
        """End call and notify server

        This object (self) is not usable after this call except that
        this method is idempotent so may be called multiple times.
        """
        if self._destroyed:
            # allows unsubscribe to be idempotent
            return
        if not self.is_subscription:
            raise AssertionError("call is not a subscription")
        try:
            self._client.unsubscribe_call(self._call_id)
        finally:
            self.destroy()

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.destroy()
        return False

    def __repr__(self):
        def _d():
            return " DESTROYED" if self._destroyed else ""

        return f"<{self.__class__.__name__} {self.api_name}#{self._call_id}{_d()}>"



from pykern import pkconfig
from pykern import pkio
import os


def _parser(value):
    if value == False:
        pkconfig.raise_error("Error prefix")

cfg = pkconfig.init(
    x=(False, _parser, "will error"),
)



# -*- coding: utf-8 -*-
"""pytest for `pykern.pkdebug.pkdformat`

:copyright: Copyright (c) 2020 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_pkdformat():
    from pykern import pkconfig

    pkconfig.reset_state_for_testing(
        {
            "PYKERN_PKDEBUG_MAX_DEPTH": "2",
            "PYKERN_PKDEBUG_MAX_ELEMENTS": "5",
            "PYKERN_PKDEBUG_MAX_STRING": "5",
        }
    )

    from pykern.pkdebug import pkdformat
    from pykern.pkunit import pkeq

    def _e(expected, value):
        pkeq(expected, pkdformat("{}", value))

    _e(
        "{'a': 'b', 'c': {'d': {<SNIP>}}, 'h': 'i'}",
        {"a": "b", "c": {"d": {"e": {"f": "g"}}}, "h": "i"},
    )
    _e(
        "{1}",
        set([1]),
    )

    _e(
        "(1, 2, 3, 4, 5)",
        (1, 2, 3, 4, 5),
    )
    _e(
        "[1, 2, 3, 4, 5, <SNIP>]",
        [1, 2, 3, 4, 5, 6],
    )
    _e(
        "(1, 2, 3, 4)",
        (1, 2, 3, 4),
    )
    _e(
        "(1, {2, 3}, {'passw<SNIP>': <REDACTED>}, [6, 7])",
        (1, {2, 3}, {"password": 5}, [6, 7]),
    )
    _e(
        "{'Secre<SNIP>': <REDACTED>, 'c2': {'botp': <REDACTED>}, 'q3': ['passw<SNIP>', 1], 's4': 'r', 't5': 'u', <SNIP>}",
        {
            "Secret1": "b",
            "c2": {"botp": "a"},
            "totp7": "iiii",
            "q3": ["password", 1],
            "x6": "y",
            "s4": "r",
            "t5": "u",
        },
    )
    _e("a" * 5 + "<SNIP>", "a" * 80)
    _e("<SNIP>" + "a" * 5, '\n  File "' + "a" * 80)

    class T:
        def pkdebug_str(self):
            return "foo"

    _e("foo", T())


def test_pkdformat_spec():
    from pykern import pkdebug, pkunit
    import decimal

    pkunit.pkeq(
        "1.23 None 003 01 True",
        pkdebug.pkdformat("{:.2f} {} {:03d} {:02d} {}", 1.23, None, 3, True, True),
    )



"""test too many failures with max_procs

:copyright: Copyright (c) 2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_restarable():
    from pykern import pkio
    from pykern import pkunit
    from pykern.pkcli import test

    # force stop on first failure
    test._cfg.max_failures = 1
    for p in (1, 2, 3):
        with pkunit.save_chdir_work() as d:
            pkunit.data_dir().join("tests").copy(d.join("tests"))
            with pkunit.pkexcept(
                f"FAILED=1 passed={2 if p > 2 else 1}",
            ):
                test.default_command(f"max_procs={p}")



# -*- coding: utf-8 -*-
"""Demonstrates RadiaSoft style testing.

Some Rules:
    We use  `pytest <http://pytest.org>`_ for unit testing.

    1. Terminology: a "test" is the file. Functions that
       begin with ``test_`` are cases.
    #. Test cases should be simple and test only one behavior
    #. Assertions messages should be informative, including
       values that are not in the assertion comparison itself.
    #. Use pytest fixtures for setup/teardown. See
       `yield fixtures <http://doc.pytest.org/en/latest/yieldfixture.html>`_
       for a way to setup/teardown each test.
    #. Use :func:`pykern.pkunit.save_chdir_work` to get automatic
       cleanup of files at start of test. The work dir stays
       around after the test runs, which helps debugging.
    #. Avoid global imports, even of the module under test.
       The pytest "collector" imports all tests files into a
       single process so you want this to go fast and not
       to initialize modules during collection.

:copyright: Copyright (c) 2016 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_ema_compute():
    from pykern import pkexample

    ema = pkexample.EMA(4)
    ema.compute(20.0)
    assert 16.0 == ema.compute(10.0), "Adding a value to the series changes the average"


def test_ema_init():
    from pykern import pkexample

    assert 5.0 == pkexample.EMA(1).compute(5.0), "First values sets initial average"


def test_ema_init_deviance():
    from pykern import pkexample

    with pytest.raises(AssertionError) as e:
        pkexample.EMA(0)
    assert "must be greater" in str(e.value)



# -*- coding: utf-8 -*-
"""conforming case_dirs

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_case_dirs_bytes():
    from pykern import pkunit

    for d in pkunit.case_dirs(group_prefix="bytes", is_bytes=True):
        # Copy does all the work, just a sanity check here
        pkunit.pkok(d.join("dot.gif").check(exists=True), "dot.gif doesn't exist")


def test_case_dirs_conformance():
    from pykern import pkunit

    for d in pkunit.case_dirs(group_prefix="conformance"):
        i = d.join("in.txt").read()
        pkunit.pkeq(d.basename + "\n", i)
        d.join("out.txt").write(i)


def test_case_dirs_deviance():
    from pykern import pkunit
    import re

    with pkunit.pkexcept("See stack above"):
        for d in pkunit.case_dirs(group_prefix="deviance"):
            with pkunit.ExceptToFile():
                d.join("in.txt").read()
    pkunit.pkre("not.*found.*deviance-1/in.txt", d.join(pkunit.PKSTACK_PATH).read())


def test_case_dirs_curly_brackets():
    from pykern import pkunit

    with pkunit.pkexcept("1c1\n< expected\n---\n> {unexpected}"):
        for d in pkunit.case_dirs(group_prefix="curly_brackets"):
            pass


def test_no_files():
    from pykern import pkunit

    with pkunit.pkexcept("no files found"):
        for d in pkunit.case_dirs(group_prefix="no_files"):
            continue



# -*- coding: utf-8 -*-
"""Test that module imports.

You should delete the test once you have real tests.
Only necessary if you have no other tests so that
tox will work.
"""
import pytest


def test_1():
    import private



from __future__ import absolute_import, division, print_function
from pykern import pkresource

from pykern import pkjinja

def render(out):
    v = {'k1': 'v1'}
    return pkjinja.render_resource('t1', v, out)



"""test pkasyncio

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

_URI = "/echo"

_PORT = None


def test_websocket():
    import os, signal, time
    from pykern import pkunit, pkdebug

    global _PORT

    _PORT = pkunit.unbound_localhost_tcp_port(37000, 38000)
    p = os.fork()
    if p == 0:
        try:
            pkdebug.pkdlog("start server")
            _server()
        except Exception as e:
            pkdebug.pkdlog("server exception={} stack={}", e, pkdebug.pkdexc())
        finally:
            os._exit(0)
    try:
        time.sleep(1)
        _client()
    finally:
        os.kill(p, signal.SIGKILL)


def _client():
    from pykern import pkdebug, pkunit
    import asyncio

    async def _all():
        w = await _open()
        pkdebug.pkdlog("open")
        await _send(w, "hello")
        pkdebug.pkdlog("send")
        pkunit.pkeq("hello", await _recv(w))
        pkdebug.pkdlog("recv")

    async def _open():
        from tornado import httpclient, websocket

        return await websocket.websocket_connect(
            httpclient.HTTPRequest(
                url=f"ws://{pkunit.LOCALHOST_IP}:{_PORT}{_URI}",
            ),
            ping_interval=100,
            ping_timeout=1000,
        )

    async def _recv(ws):
        return await ws.read_message()

    async def _send(ws, text):
        await ws.write_message(text)

    asyncio.run(_all())

    # This does not work. The read_message hangs.
    # try:
    #     w = asyncio.run(_open())
    #     pkdebug.pkdlog("open")
    #     asyncio.run(_send(w))
    #     pkdebug.pkdlog("send")
    #     pkdebug.pkdlog("recv={}", asyncio.run(_recv(w)))
    # except Exception as e:
    #     pkdebug.pkdlog("exception={} stack={}", e, pkdebug.pkdexc())


def _server():
    from pykern import pkasyncio, pkdebug, pkunit
    from pykern.pkcollections import PKDict
    from tornado import websocket

    class _Echo(websocket.WebSocketHandler):
        def open(self):
            # just to print something
            pkdebug.pkdlog("remote_ip={}", self.request.remote_ip)

        async def on_message(self, msg):
            import asyncio

            try:
                pkdebug.pkdlog(msg)
                self.write_message(msg)
                pkdebug.pkdlog("send")
            except Exception as e:
                pkdebug.pkdlog("exception {}", e)

    l = pkasyncio.Loop()
    l.http_server(
        PKDict(
            uri_map=((_URI, _Echo),),
            tcp_port=_PORT,
            tcp_ip=pkunit.LOCALHOST_IP,
        )
    )
    l.start()



# -*- coding: utf-8 -*-
"""pkcli.test test

:copyright: Copyright (c) 2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_cases(capsys):
    from pykern import pkunit, pkio
    from pykern.pkcli import test
    from pykern.pkcollections import PKDict
    import re

    # Wait for #410 which would support this
    for d in pkunit.case_dirs():
        try:
            test.default_command(".")
        except Exception:
            pass
        o, e = capsys.readouterr()
        pkio.write_text("stdout.txt", o)
        pkio.write_text("stderr.txt", e)
        pkunit.pkre("Error: .+ coroutine .+ was never awaited", o)



# -*- coding: utf-8 -*-
"""Declarative module configuration with dynamic value injection

Module Declaration
------------------

Modules declare their configuration via `init`. Here is how `pkdebug`
declares its config params::

    cfg = pkconfig.init(
        control=(None, re.compile, 'Pattern to match against pkdc messages'),
        want_pid_time=(False, bool, 'Display pid and time in messages'),
        output=(None, _cfg_output, 'Where to write messages either as a "writable" or file name'),
    )

A param tuple contains three values:

    0. Default value, in the expected type
    1. Callable that can convert a string or other type into a valid value
    2. A docstring briefly explaining the configuration element

The returned ``cfg`` object is ready to use after the call. It will contain
the config params as defined or an exception will be raised.

Config Values
-------------

Configuration is returned as nested dicts. The values themselves could
be any Python object. In this case, we have a string and a file object
for the two parameters. We called `os.getcwd` and referred to
`sys.stdout` in param values.

Summary
-------

Here are the steps to configuring an application:

1. When the first module calls `init`, pkconfig gets environment variables
   to create a single dict of param values, unparsed.

2. `init` looks for the module's params in the unparsed values.

3. If the parameter is found, that value is used. Else, the default is merged
   into the dict and used.

4. The parameter value is then resolved with `str.format`. If the value
   is a `list` it will be joined with any previous value (e.g. default).

5. The resolved value is parsed using the param's declared ``parser``.

6. The result is stored in the merged config and also stored in the module's
   `Params` object .

7. Once all params have been parsed for the module, `init` returns the `Params`
   object to the module, which can then use those params to initialize itself.

:copyright: Copyright (c) 2015-2019 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html

"""
from __future__ import absolute_import, division, print_function

# Import the minimum number of modules and none from pykern
# pkconfig is the first module imported by all other modules in pykern
import collections
import copy
import importlib
import inspect
import os
import re
import sys

# These modules have very limited imports to avoid loops with config imports
from pykern.pkcollections import PKDict
from pykern import pkconst
from pykern import pkinspect

#: Python version independent value of string instance check
STRING_TYPES = pkconst.STRING_TYPES

#: Environment variable holding channel (defaults to "dev")
CHANNEL_ATTR = "pykern_pkconfig_channel"

_KEY_PATTERN = "^[A-Z][A-Z0-9_]*[A-Z0-9]$"

#: Validate environ key: Must be upper case, cannot begin with non-letter or end with an underscore
_ENV_KEY_RE = re.compile(_KEY_PATTERN)

#: Validate key: Cannot begin with non-letter or end with an underscore
KEY_RE = re.compile(_KEY_PATTERN, flags=re.IGNORECASE)

#: parse_tuple splits strings on this
TUPLE_SEP = ":"

#: Order of channels from least to most stable
VALID_CHANNELS = ("dev", "alpha", "beta", "prod")

#: Channels which can have more verbose output from the server
INTERNAL_TEST_CHANNELS = VALID_CHANNELS[0:2]

#: Configuration for this module: channel
cfg = None

#: Initialized channel (same as cfg.channel)
CHANNEL_DEFAULT = VALID_CHANNELS[0]

#: Attribute to detect parser which can parse None
_PARSE_NONE_ATTR = "pykern_pkconfig_parse_none"

#: Value to add to os.environ (see `reset_state_for_testing`)
_add_to_environ = None

#: All values in environ and add_to_environ
_raw_values = None

#: All values parsed via init()
_parsed_values = None

#: Regex used by `parse_seconds`
_PARSE_SECONDS = re.compile(
    r"^(?:(\d+)d)?(?:(?:(?:(\d+):)?(\d+):)?(\d+))?$",
    flags=re.IGNORECASE,
)

#: Regex used by `parse_bytes`
_PARSE_BYTES = re.compile(r"^(\d+)([kmgtp]?)b?$", flags=re.IGNORECASE)

#: multiplier used for qualifier on `parse_bytes`
_PARSE_BYTES_MULTIPLIER = PKDict(
    k=1024,
    m=1024**2,
    g=1024**3,
    t=1024**4,
)


class ReplacedBy(tuple, object):
    """Container for a required parameter declaration.

    Example::

        cfg = pkconfig.init(
            gone=pkconfig.ReplacedBy('pykern.pkexample.foo'),
        )

    Args:
        new_name: name of new config parameter
    """

    @staticmethod
    def __new__(cls, new_name):
        msg = "replaced by name=${}".format(new_name.upper().replace(".", "_"))
        return super(ReplacedBy, cls).__new__(
            cls,
            (None, lambda x: raise_error(msg), msg),
        )


class Required(tuple, object):
    """Container for a required parameter declaration.

    Example::

        cfg = pkconfig.init(
            any_param=(1, int, 'A parameter with a default'),
            needed=pkconfig.Required(int, 'A parameter without a default'),
        )

    Args:
        converter (callable): how to string to internal value
        docstring (str): description of parameter
    """

    @staticmethod
    def __new__(cls, *args):
        assert len(args) == 2, "{}: len(args)!=2".format(args)
        return super(Required, cls).__new__(cls, (None,) + args)


class RequiredUnlessDev(tuple, object):
    """Container for a required parameter declaration only `in_dev_mode`.

    Only required if `in_dev_mode` is true.

    Example::

        cfg = pkconfig.init(
            maybe_needed=pkconfig.RequiredUnlessDev('dev default', str, 'A parameter with a default'),
        )

    Args:
        dev_default (object): value compatible with type
        converter (callable): how to string to internal value
        docstring (str): description of parameter
    """

    @staticmethod
    def __new__(cls, *args):
        assert len(args) == 3, "{}: len(args)!=3".format(args)
        if in_dev_mode():
            return args
        return Required(args[1], args[2])


def append_load_path(load_path):
    """DEPRECATED"""
    pass


def channel_in(*args, **kwargs):
    """Test against configured channel

    Args:
        args (str): list of channels to valid
        channel (str): channel to test (default: [cfg.channel])

    Returns:
        bool: True if current channel in ``args``
    """
    if not cfg:
        _coalesce_values()
    res = False
    to_test = cfg.channel
    if kwargs and kwargs["channel"]:
        to_test = kwargs["channel"]
        assert to_test in VALID_CHANNELS, "{}: invalid channel keyword arg".format(
            to_test
        )
    for a in args:
        assert a in VALID_CHANNELS, "{}: invalid channel to test".format(a)
        if a == to_test:
            res = True
    return res


def channel_in_internal_test(channel=None):
    """Is this a internal test channel?

    Args:
        channel (str): channel to test (default: [cfg.channel])

    Returns:
        bool: True if current channel in (alpha, dev)
    """
    return channel_in(*INTERNAL_TEST_CHANNELS, channel=channel)


def init(**kwargs):
    """Declares and initializes config params for calling module.

    Args:
        kwargs (dict): param name to (default, parser, docstring)

    Returns:
        Params: `PKDict` populated with param values
    """
    if "_caller_module" in kwargs:
        # Internal use only: _values() calls init() to initialize pkconfig.cfg
        m = kwargs["_caller_module"]
        del kwargs["_caller_module"]
    else:
        if pkinspect.is_caller_main():
            pkconst.builtin_print(
                "pkconfig.init() called from __main__; cannot configure, ignoring",
                file=sys.stderr,
            )
            return None
        m = pkinspect.caller_module()
    mnp = m.__name__.split(".")
    for k in reversed(mnp):
        kwargs = {k: kwargs}
    decls = {}
    _flatten_keys([], kwargs, decls)
    _coalesce_values()
    res = PKDict()
    _iter_decls(decls, res)
    for k in mnp:
        res = res[k]
    return res


def in_dev_mode():
    """Turn on developer features

    Returns:
        bool: value of `pykern.pykern.cfg.dev_mode`

    """
    if not cfg:
        _coalesce_values()
    return cfg.dev_mode


def flatten_values(base, new):
    """Merge flattened values into base

    Keys are made all lowercase.

    Lists instances are prepended and not recursively merged.

    Args:
        base (object): dict-like that is already flattened
        new (object): dict-like that will be flattened and overriden
    Returns:
        dict: modified `base`
    """
    new_values = {}
    _flatten_keys([], new, new_values)
    # TODO(robnagler) Verify that a value x_y_z isn't set when x_y
    # exists already as a None. The other way is ok, because it
    # clears the value unless of course it's not a dict
    # then it would be a type collision
    for k in sorted(new_values.keys()):
        n = new_values[k]
        if k in base:
            b = base[k]
            if isinstance(b, list) or isinstance(n, list):
                if b is None or n is None:
                    pass
                elif isinstance(b, list) and isinstance(n, list):
                    n.extend(b)
                else:
                    raise_error(
                        "{}: type mismatch between new value ({}) and base ({})".format(
                            k.msg, n, b
                        ),
                    )
        base[k] = n
    return base


def parse_none(func):
    """Decorator for a parser which can parse None

    Args:
        callable: function to be decorated

    Returns:
        callable: func with attr indicating it can parse None
    """
    setattr(func, _PARSE_NONE_ATTR, True)
    return func


@parse_none
def parse_bool(value):
    """Default parser for `bool` types

    When the parser is `bool`, it will be replaced with this routine,
    which parses strings and None specially. `bool` values
    cannot be defaulted to `None`. They must be True or False.

    String values which return true: t, true, y, yes, 1.

    False values: f, false, n, no, 0, '', None

    Other values are parsed by `bool`.

    Args:
        value (object): to be parsed

    Returns:
        bool: True or False
    """
    if value is None:
        return False
    if isinstance(value, bool):
        return value
    if not isinstance(value, STRING_TYPES):
        return bool(value)
    v = value.lower()
    if v in ("t", "true", "y", "yes", "1"):
        return True
    if v in ("f", "false", "n", "no", "0", ""):
        return False
    raise_error("unknown boolean value={}".format(value))


def parse_bytes(value):
    """Parse bytes in `int` or n[KMGT]B? formats

    Args:
        value (object): to be parsed

    Returns:
        int: non-negative number of bytes
    """
    if isinstance(value, int):
        if value < 0:
            raise_error("bytes may not be negative value={}".format(value))
        return value
    if not isinstance(value, str):
        raise_error("bytes must be int or str value={}".format(value))
    m = _PARSE_BYTES.search(value)
    if not m:
        raise_error("bytes must match n[KMGT]B? value={}".format(value))
    v = int(m.group(1))
    x = m.group(2)
    if x:
        v *= _PARSE_BYTES_MULTIPLIER[x.lower()]
    return v


def parse_positive_int(value):
    """Parse is a positive integer

    Args:
        value (object): to be parsed

    Returns:
        int: positive value
    """
    if isinstance(value, str):
        try:
            value = int(value)
        except Exception as e:
            raise_error(f"value={value} not an integer; exception={e}")
    elif not isinstance(value, int):
        raise_error(f"value={value} must be int or str")
    if value <= 0:
        raise_error(f"value={value} must be positive")
    return value


def parse_seconds(value):
    """Parse seconds in `int` or DdH:M:S formats

    Args:
        value (object): to be parsed

    Returns:
        int: non-negative number of seconds
    """
    if isinstance(value, int):
        if value < 0:
            raise_error(f"seconds may not be negative value={value}")
        return value
    if not isinstance(value, str):
        raise_error(f"seconds must be int or str value={value}")
    m = _PARSE_SECONDS.search(value)
    if not m or not any(m.groups()):
        raise_error(f"seconds must match [Dd][[[H:]M:]S] value={value}")
    v = 0
    for x, i in zip((86400, 3600, 60, 1), m.groups()):
        if i is not None:
            v += int(i) * x
    return v


#: deprecated version of parse_seconds
parse_secs = parse_seconds


@parse_none
def parse_set(value):
    """Default parser for `set` and `frozenset` types

    When the parser is `set` or `frozenset`, it will be replaced with this routine,
    which parses strings, lists, and sets. It splits strings on ':'.

    Args:
        value (object): to be parsed

    Returns:
        frozenset: may be an empty set
    """
    return frozenset(parse_tuple(value))


@parse_none
def parse_tuple(value):
    """Default parser for `tuple` types

    When the parser is `tuple`, it will be replaced with this routine,
    which parses strings, lists, sets, and tuples. It splits strings on ':'.

    Args:
        value (object): to be parsed

    Returns:
        tuple: may be an empty tuple

    """
    if value is None:
        return tuple()
    if isinstance(value, tuple):
        return value
    if isinstance(value, (list, set, frozenset)):
        return tuple(value)
    assert isinstance(
        value, STRING_TYPES
    ), "unable to convert type={} to tuple; value={}".format(type(value), value)
    return tuple(value.split(TUPLE_SEP))


def raise_error(msg):
    """Call when there is a config problem"""
    raise AssertionError(msg)


def reset_state_for_testing(add_to_environ=None):
    """Clear the raw values and append add_to_environ

    Only used for unit tests. ``add_to_environ`` overrides previous
    value.

    Args:
        add_to_environ (dict): values to augment to os.environ
    """
    global _raw_values, _add_to_environ
    _raw_values = None
    _add_to_environ = copy.deepcopy(add_to_environ)


def to_environ(cfg_keys, values=None, exclude_re=None):
    """Export config (key, values) as dict for environ

    cfg_keys is a list of dotted words (``['pykern.pkdebug.control']``).

    Simple globs (``pykern.pkdebug.*``) are supported, which match
    any word character. This can be used to ensure there is enough depth,
    e.g. pykern.*.* means there must be at least two undercores in
    the environment variable.

    Only environ and add_to_environ config will be considered, not
    default values, which will be assumed to be processed the same
    way in a subprocess using this environ.

    Args:
        cfg_keys (iter): keys to find values for
        values (mapping): use for configuration to parse [actual config]
        exclude_re (object): compile re or str to ignore matches
    Returns:
        PKDict: keys and values (str)
    """
    c = flatten_values({}, values) if values else _coalesce_values()
    res = PKDict()

    if exclude_re and isinstance(exclude_re, STRING_TYPES):
        exclude_re = re.compile(exclude_re, flags=re.IGNORECASE)

    def a(k, v):
        if exclude_re and exclude_re.search(k):
            return
        if not isinstance(v, STRING_TYPES):
            if v is None:
                v = ""
            elif isinstance(v, bool):
                v = "1" if v else ""
            elif isinstance(v, (frozenset, list, set, tuple)):
                v = TUPLE_SEP.join(v)
            else:
                v = str(v)
        res[k.upper()] = v

    for k in cfg_keys:
        k = k.lower().replace(".", "_")
        if "*" not in k:
            if k in c:
                a(k, c[k])
            continue
        r = re.compile(k.replace("*", r"\w+"), flags=re.IGNORECASE)
        for x, v in c.items():
            if r.search(x):
                a(x, v)
    return res


class _Declaration(object):
    """Initialize a single parameter declaration

    Args:
        name (str): for error output
        value (tuple or dict): specification for parameter

    Attributes:
        default (object): value to be assigned if not explicitly configured
        docstring (str): documentation for the parameter
        group (Group): None or Group instance
        parser (callable): how to parse a configured value
        required (bool): the param must be explicitly configured
    """

    def __init__(self, value):
        if isinstance(value, dict):
            self.group = value
            self.parser = None
            self.default = None
            self.docstring = ""
            # TODO(robnagler) _group_has_required(value)
            self.required = False
            return
        assert len(value) == 3, "{}: declaration must be a 3-tuple".format(value)
        self.default = value[0]
        self.parser = value[1]
        self.docstring = value[2]
        assert callable(self.parser), "{}: parser must be a callable: ".format(
            self.parser, self.docstring
        )
        self.group = None
        self.required = isinstance(value, Required)
        self._fixup_parser()

    def _fixup_parser(self):
        if self.parser == bool:
            t = (int,)
            self.parser = parse_bool
        elif self.parser == tuple:
            t = (tuple,)
            self.parser = parse_tuple
        elif self.parser in (set, frozenset):
            t = (frozenset, set, tuple)
            self.parser = parse_set
        else:
            return
        if self.required:
            return
        # better error message than what parser might put out
        assert isinstance(
            self.default, t
        ), "default={} must be a type={} docstring={}".format(
            self.default,
            [str(x) for x in t],
            self.docstring,
        )
        # validate the default
        self.default = self.parser(self.default)


class _Key(str, object):
    """Internal representation of a key for a value

    The str value is lowercase joined with ``_``. For debugging,
    ``msg`` is printed (original case, joined on '.'). The parts
    are saved for creating nested values.
    """

    @staticmethod
    def __new__(cls, parts):
        self = super(_Key, cls).__new__(cls, "_".join(parts).lower())
        self.parts = parts
        self.msg = ".".join(parts)
        return self


def _clean_environ():
    """Ensure os.environ keys are valid (no bash function names)

    Also sets empty string to `None`.

    Returns:
        dict: copy of a cleaned up `os.environ`
    """

    def _value(env, k):
        return env[k] if len(env[k]) > 0 else None

    def _add_others(env, res):
        for k in sorted(env.keys()):
            if KEY_RE.search(k):
                x = k.upper()
                if x not in res:
                    res[x] = _value(env, k)

    def _add_uppers(env, res):
        for k in sorted(env.keys()):
            if _ENV_KEY_RE.search(k):
                res[k] = _value(env, k)
                del env[k]

    res = {}
    env = os.environ.copy()
    # TODO(robnagler) this makes it easier to set debugging, but it's a hack
    if "pkdebug" in env and "PYKERN_PKDEBUG_CONTROL" not in env:
        env["PYKERN_PKDEBUG_CONTROL"] = env["pkdebug"]
    if _add_to_environ:
        env.update(_add_to_environ)
    _add_uppers(env, res)
    _add_others(env, res)
    return res


def _coalesce_values():
    """Coalesce environ and add_to_environ

    Sets up channel.

    Returns:
        dict: raw values
    """
    global _raw_values, _parsed_values
    global cfg
    if _raw_values:
        return _raw_values
    values = {}
    env = _clean_environ()
    flatten_values(values, env)
    channel = values.get(CHANNEL_ATTR, CHANNEL_DEFAULT)
    assert channel in VALID_CHANNELS, "{}: invalid ${}; must be {}".format(
        channel, CHANNEL_ATTR.upper(), VALID_CHANNELS
    )
    values[CHANNEL_ATTR] = channel
    _raw_values = values
    _parsed_values = dict(((_Key([k]), v) for k, v in env.items()))
    a = PKDict(
        _caller_module=sys.modules[__name__],
        channel=Required(str, "which (stage) function returns config"),
    )
    cfg = init(**a)
    a.dev_mode = (channel_in("dev"), bool, "controls development features")
    cfg = init(**a)
    return _raw_values


def _flatten_keys(key_parts, values, res):
    """Turns values into non-nested dict with `_Key` keys, flat

    Args:
        key_parts (list): call with ``[]``
        values (dict): nested dicts of config values
        res (dict): result container (call with ``{}``)
    """
    for k in values:
        v = values[k]
        k = _Key(key_parts + k.split("."))
        assert KEY_RE.search(k), "{}: invalid key must match {}".format(k.msg, KEY_RE)
        assert not k in res, "{}: duplicate key".format(k.msg)
        if isinstance(v, dict):
            _flatten_keys(k.parts, v, res)
        else:
            # Only store leaves
            res[k] = v


def _iter_decls(decls, res):
    """Iterates decls and resolves values into res

    Args:
        decls (dict): nested dictionary of a module's cfg values
        res (PKDict): result configuration for module
    """
    for k in sorted(decls.keys()):
        # TODO(robnagler) deal with keys with '.' in them (not possible?)
        d = _Declaration(decls[k])
        r = res
        for kp in k.parts[:-1]:
            if kp not in r:
                r[kp] = PKDict()
            r = r[kp]
        kp = k.parts[-1]
        if d.group:
            r[kp] = PKDict()
            continue
        try:
            r[kp] = _resolver(d)(k, d)
        except Exception as e:
            pkinspect.append_exception_reason(e, f"key={kp}, value={decls[k]}")
            raise
        _parsed_values[k] = r[kp]


def _resolver(decl):
    """How to resolve values for declaration

    Args:
        decl (_Declaration): what to resolve

    Returns:
        callable: `_resolve_dict`, `_resolve_list`, or `_resolve_value`
    """
    if dict == decl.parser:
        return _resolve_dict
    if list == decl.parser:
        return _resolve_list
    return _resolve_value


def _resolve_dict(key, decl):
    # TODO(robnagler) assert "required"
    res = PKDict(copy.deepcopy(decl.default) if decl.default else {})
    assert isinstance(res, dict), "{}: default ({}) must be a dict".format(
        key.msg, decl.default
    )
    key_prefix = key + "_"
    for k in reversed(sorted(_raw_values.keys())):
        if k != key and not k.startswith(key_prefix):
            continue
        r = res
        if len(k.parts) == 1:
            # os.environ has only one part (no way to split on '.')
            # so we have to assign the key's suffix manually
            ki = k.parts[0][len(key_prefix) :]
            # TODO(robnagler) if key exists, preserve case (only for environ)
        else:
            kp = k.parts[len(key.parts) : -1]
            for k2 in kp:
                if not k2 in r:
                    r[k2] = PKDict()
                else:
                    assert isinstance(
                        r[k2], dict
                    ), "{}: type collision on existing non-dict ({}={})".format(
                        k.msg, k2, r[k2]
                    )
                r = r[k2]
            ki = k.parts[-1]
        r[ki] = _raw_values[k]
    return res


def _resolve_list(key, decl):
    # TODO(robnagler) assert required
    res = copy.deepcopy(decl.default) if decl.default else []
    assert isinstance(res, list), "{}: default ({}) must be a list".format(
        key.msg, decl.default
    )
    if key not in _raw_values:
        assert not decl.required, "{}: config value missing and is required".format(k)
        return res
    if not isinstance(_raw_values[key], list):
        if _raw_values[key] is None:
            return None
        raise_error(
            "{}: value ({}) must be a list or None".format(key.msg, _raw_values[key]),
        )
    return _raw_values[key] + res


def _resolve_value(key, decl):
    if key in _raw_values:
        res = _raw_values[key]
    else:
        assert not decl.required, "{}: config value missing and is required".format(
            key.msg
        )
        res = decl.default
    # TODO(robnagler) FOO_BAR='' will not be evaluated. It may need to be
    # if None is not a valid option and there is a default
    if res is None and not hasattr(decl.parser, _PARSE_NONE_ATTR):
        return None
    return decl.parser(res)


def _z(msg):
    """Useful for debugging this module"""
    with open("/dev/tty", "w") as f:
        f.write(str(msg) + "\n")



"""Support for `pykern.api` tests

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# Defer as many pykern imports as possible to defer pkconfig runing
from pykern.pkcollections import PKDict
import os
import signal
import time

_AUTH_TOKEN = "http_unit_auth_secret"


class Setup:
    """Usage::

        async with http_unit.Setup(api_classes=(_class())) as c:
            from pykern.pkcollections import PKDict
            from pykern import pkunit

            e = PKDict(ping="pong")
            pkunit.pkeq(e.pkupdate(counter=1), await c.call_api("echo", e))
            pkunit.pkeq(e.pkupdate(counter=2), await c.call_api("echo", e))

    May be subclassed to start multiple servers.
    """

    def __init__(self, **server_config):
        # Must be first
        self._global_config()
        self.http_config = self._http_config()
        self.server_config = self._server_config(server_config)
        self.server_pid = self._server_process()
        time.sleep(2)
        self.client = self._client()

    def destroy(self):
        """Destroy client and kill attributes ``*_pid`` attrs"""

        self.client.destroy()
        for p in filter(lambda x: x.endswith("_pid"), dir(self)):
            try:
                os.kill(getattr(self, p), signal.SIGKILL)
            except Exception:
                pass

    def _client(self):
        """Creates a client to be used for requests.

        Called in `__init__`.

        Returns:
            object: http client, set to ``self.client``
        """
        from pykern.api import client

        return client.Client(self.http_config.copy())

    def _client_awaitable(self):
        """How to connect to client

        Awaited in `__aenter__`.

        Returns:
            Awaitable: coroutine to connect to client
        """
        return self.client.connect(PKDict(token=_AUTH_TOKEN))

    def _global_config(self, **kwargs):
        """Initializes os.environ and pkconfig

        Called first.

        Args:
            kwargs (dict): merged into environ and config (from subclasses)
        """

        c = PKDict(
            PYKERN_PKDEBUG_WANT_PID_TIME="1",
            **kwargs,
        )
        os.environ.update(**c)
        from pykern import pkconfig

        pkconfig.reset_state_for_testing(c)

    def _http_config(self):
        """Initializes ``self.http_config``

        Returns:
            PKDict: configuration to be shared with client and server
        """
        from pykern import pkconst, util

        return PKDict(
            # any uri is fine
            api_uri="/api_unit",
            # just needs to be >= 16 word (required by http) chars; apps should generate this randomly
            tcp_ip=pkconst.LOCALHOST_IP,
            tcp_port=util.unbound_localhost_tcp_port(),
        )

    def _server_config(self, init_config):
        """Config to be passed to `pykern.api.server.start` in `server_start`

        A simple `pykern.api.auth_api.AuthAPI` implementation is defaulted if not in ``init_config.api_classes``.

        Args:
            init_config (dict): what was passed to `__init__`
        Returns:
            PKDict: configuration for `server_start`
        """

        def _api_classes(init_classes):
            from pykern import pkdebug
            from pykern.api import auth_api

            class AuthAPI(auth_api.AuthAPI):

                def token(self):
                    return _AUTH_TOKEN

            return init_classes + [AuthAPI]

        rv = PKDict(init_config) if init_config else PKDict()
        rv.pksetdefault(
            api_classes=(),
            attr_classes=(),
            coros=(),
            http_config=PKDict,
        )
        rv.http_config.pksetdefault(**self.http_config)
        rv.api_classes = _api_classes(list(rv.api_classes))
        return rv

    def _server_process(self):
        """Call `server_start` in separate process

        Override this method to start multiple servers, saving pids in
        attributes that end in ``_pid`` so that `destroy` will kill
        them.

        Returns:
            int: pid of process

        """
        from pykern import pkdebug

        if rv := os.fork():
            return rv
        try:
            pkdebug.pkdlog("start server")
            self._server_start()
        except Exception as e:
            pkdebug.pkdlog("exception={} stack={}", e, pkdebug.pkdexc())
        finally:
            os._exit(0)

    def _server_start(self):
        """Calls api.server.start with ``self.server_config``"""
        from pykern.api import server

        server.start(**self.server_config)

    async def __aenter__(self):
        await self._client_awaitable()
        return self.client

    async def __aexit__(self, *args, **kwargs):
        self.destroy()
        return False



"""API wrapper

:copyright: Copyright (c) 2024 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp, pkdformat
import contextlib

_SUBSCRIPTION_ATTR = "pyern_quest_subscription"


class API(PKDict):
    """Holds request context for all API calls."""

    METHOD_PREFIX = "api_"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._destroyed = False

    def is_quest_end(self):
        return self._destroyed

    def quest_end(self, in_error=True):
        def _attr_end(name, attr):
            try:
                attr.quest_end(self, in_error=in_error)
            except Exception:
                pkdlog("destroy failed {}={} stack={}", name, attr, pkdexc())

        if self._destroyed:
            return
        x = reversed(list(self.__attrs()))
        self.clear()
        # must be after the clear
        self._destroyed = True
        for k, v in x:
            _attr_end(k, v)

    def quest_init(self, attr_classes, init_kwargs):
        def _set(name, attr):
            if not isinstance(attr, Attr):
                raise AssertionError(f"type=type(obj) not Attr name={name}")
            if name in self:
                raise AssertionError(f"name={name} already added")
            self[name] = attr

        for a in attr_classes:
            s = a if isinstance(a, Attr) else a.quest_init(self, init_kwargs)
            _set(s.ATTR_KEY, s)

    def quest_start(self):
        for _, v in self.__attrs():
            v.quest_start(self)

    def __attrs(self):
        for k, v in self.items():
            if isinstance(k, Attr):
                yield k, v


class Attr(PKDict):
    #: shared Attrs do not have link to qcall
    IS_SINGLETON = False

    def __init__(self, qcall, **kwargs):
        """Initialize object

        Subclasses must define ATTR_KEY so it can be added to qcall.

        If `IS_SINGLETON` is true then qcall must be None. This will
        only be called outside of quest_init. Otherwise, qcall is
        bound to instance.

        Args:
            qcall (API): what qcall is being initialized
            kwargs (dict): inserted into dictionary

        """
        if self.IS_SINGLETON:
            assert qcall is None
            super().__init__(**kwargs)
        else:
            # It may be qcall is None at this point, but that's ok. Will be set below
            super().__init__(qcall=qcall, **kwargs)

    def quest_end(self, qcall, in_error):
        """Called when quest ends

        Right before destroy. No other attributes are available.

        Args:
            qcall (API): qcall being ended
            in_error (bool): True, aborting quest. False, successful quest [True]
        """
        pass

    @classmethod
    def quest_init(cls, qcall, init_kwargs):
        """Initialize an instance of cls and put on qcall

        If `IS_SINGLETON`, qcall is not put on self. `kwargs` must contain ATTR_KEY,
        which is an instance of class.

        Args:
            qcall (API): quest being initialized
            init_kwargs (PKDict): values to passed to `start`
        Returns:
            Attr: instance to bind to quest
        """
        if not cls.IS_SINGLETON:
            self = cls(qcall, **init_kwargs)
        elif (self := init_kwargs.get(cls.ATTR_KEY)) is None:
            raise AssertionError(
                f"init_kwargs does not contain singleton key={cls.ATTR_KEY}"
            )
        if not isinstance(self, Attr):
            raise AssertionError(f"{cls.ATTR_KEY}={self} not instance of Attr")
        return self

    def quest_start(self, qcall):
        """Called after all attrs are initialized

        Args:
            qcall (API): quest being started
        """
        pass


class Spec:
    # qspec
    pass


@contextlib.contextmanager
def start(api_class, attr_classes, **kwargs):
    qcall = api_class()
    e = True
    try:
        qcall.quest_init(attr_classes, PKDict(kwargs))
        qcall.quest_start()
        yield qcall
        e = False
    finally:
        qcall.quest_end(in_error=e)



"""Wrapper for subprocess.

:copyright: Copyright (c) 2016 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkdebug import pkdc, pkdexc, pkdp, pkdlog
import os
import signal
import six
import subprocess
import threading

#: Caught signals
_SIGNALS = (signal.SIGTERM, signal.SIGINT)


def check_call_with_signals(cmd, output=None, env=None, msg=None, recursive_kill=False):
    """Run cmd, writing to output.

    stdin is `os.devnull`.

    Passes SIGTERM and SIGINT on to the child process. If `output`
    is a string, it will be opened in write ('w') mode.

    Args:
        cmd (list): passed to subprocess verbatim
        output (file or str): where to write stdout and stderr
        env (dict): environment to use
        recursive_kill (bool): EXPERIMENTAL: kill all process children, recursively
    """
    assert (
        _is_main_thread()
    ), "subprocesses which require signals need to be started in main thread"
    p = None
    prev_signal = dict([(sig, signal.getsignal(sig)) for sig in _SIGNALS])
    pid = None
    all_pids = set()

    def signal_handler(sig, frame):
        try:
            if p:
                p.send_signal(sig)
        except Exception:
            # Nothing we can do, still want to cascade
            pass
        finally:
            ps = prev_signal[sig]
            if ps in (None, signal.SIG_IGN, signal.SIG_DFL):
                return
            ps(sig, frame)

    def wait_pid():
        """Iteratively and recursively gather all children

        mpiexec sets a session, and doesn't cascade signals so processes
        can stay running after an exit.
        """
        # always SIGKILL the process we started
        all_pids.add(pid)
        if not recursive_kill:
            # simple process running
            return p.wait()

        import psutil, time

        # EXPERIMENTAL
        z = psutil.Process(pid)
        t = 0.1
        while True:
            all_pids.update(
                (c.pid for c in z.children(recursive=True)),
            )
            x, s = os.waitpid(pid, os.WNOHANG)
            if x != 0:
                break
            time.sleep(t)
            # first sleep is very fast, just in case a fast
            # process starts. Then polling less frequently
            # helps avoid thrashing, especially with mpi.
            t = 0.5
        return s

    try:
        stdout = output
        if isinstance(output, six.string_types):
            stdout = open(output, "w")
        stderr = subprocess.STDOUT if stdout else None
        for sig in _SIGNALS:
            signal.signal(sig, signal_handler)
        p = subprocess.Popen(
            cmd,
            stdin=open(os.devnull),
            stdout=stdout,
            stderr=stderr,
            env=env,
        )
        pid = p.pid
        if msg:
            msg("{}: started: {}", pid, cmd)
        s = wait_pid()
        p = None
        if s != 0:
            raise RuntimeError("error exit({})".format(s))
        if msg:
            msg("{}: normal exit(0): {}", pid, cmd)
    except Exception as e:
        if msg:
            msg("{}: exception: {} {}", pid, cmd, pkdexc())
        raise
    finally:
        for sig in _SIGNALS:
            signal.signal(sig, prev_signal[sig])
        if p is not None:
            if msg:
                msg("{}: terminating: {}", pid, cmd)
            try:
                p.terminate()
                time.sleep(0.1)
            except Exception:
                pass
        for x in all_pids:
            try:
                os.kill(x, signal.SIGKILL)
                # maybe we didn't catch all the children so try this
                os.killpg(x, signal.SIGKILL)
            except Exception:
                pass
        if stdout != output:
            stdout.close()


def _is_main_thread():
    """Need to determine if the main thread for setting signals

    Returns:
        bool: if running in the main thread
    """
    if hasattr(threading, "main_thread"):
        # Python 3
        return threading.current_thread() == threading.main_thread()
    # Python 2: See http://stackoverflow.com/a/23207116
    return threading.current_thread().__class__ == threading._MainThread



def args_and_imports(
    be,
    a,
    very,
    poorly,
    formatted,
    function,
    that,
    noone,
    should,
    ever,
    use,
    it,
    has,
    way,
    too,
    many,
    arguments,
):
    from pykern.pkdebug import pkdp

    x = "abcdefg"

    def foo():
        return 1

    def bar():
        return 3



"""Support routines, including run dir resolution.

:copyright: Copyright (c) 2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# Root module: avoid global pykern imports
import os.path
import sys

_ACCEPTABLE_CONTROL_CODE_RATIO = 0.33
_DEFAULT_ROOT = "run"
_DEV_ONLY_FILES = ("setup.py", "pyproject.toml")
_VALID_ASCII_CONTROL_CODES = frozenset((0x7, 0x8, 0x9, 0xA, 0xB, 0xC, 0xD, 0x1B))

_dev_run_dir = None


class APIError(Exception):
    """Application level errors or exceptions sent to client and raised on client"""

    def __init__(self, fmt, *args, **kwargs):
        from pykern.pkdebug import pkdformat

        super().__init__(pkdformat(fmt, *args, **kwargs) if args or kwargs else fmt)


def cfg_absolute_dir(value):
    """Parser function for absolute dir config value

    Args:
        value (str): string absolute path to dir

    Returns:
        py.path.Local absolute path to dir
    """
    from pykern import pkio, pkconfig

    if not os.path.isabs(value):
        pkconfig.raise_error("must be absolute")
    if not os.path.isdir(value):
        pkconfig.raise_error("must be a directory and exist")
    return pkio.py_path(value)


def dev_run_dir(package_object):
    """Resolves run directory if in development mode and creates it if it doesn't exist.

    Note that this assumes a file structure like pykern's where <project_name>'s __init__.py
    lives in <project_name>/<project_name>/

    Args:
        package_object (object): A python object whose root package we are resolving the run directory for

    Returns:
        py.path.local: The absolute path to the run directory.

    Raises:
        AssertionError: Raised when not in development mode
    """
    global _dev_run_dir

    from pykern import pkconfig

    def _check_dev_files(root):
        """Check for files that only exist in development"""
        return any([root.join(f).check() for f in _DEV_ONLY_FILES])

    def _mkdir():
        from pykern import pkinspect, pkio, pkunit

        if pkunit.is_test_run():
            return pkunit.work_dir()
        r = (
            pkio.py_path(
                sys.modules[pkinspect.root_package(package_object)].__file__,
            )
            .dirpath()
            .dirpath()
        )
        if not _check_dev_files(r):
            # Don't run from an install directory
            r = pkio.py_path()
        return pkio.mkdir_parent(r.join(_DEFAULT_ROOT))

    if _dev_run_dir:
        return _dev_run_dir
    if not pkconfig.in_dev_mode():
        raise AssertionError("run dir root must be configured except in dev")
    _dev_run_dir = _mkdir()
    return _dev_run_dir


def is_pure_text(value, is_truncated=False):
    """Guesses if value is text data using heuristics:

    Checks if value can be utf-8 decoded.

    If fails to decode and is_truncated, probes backwards up to 4 chars
    (4 is maximum length of a utf-8 char) in case a valid utf-8
    char was truncated at the boundary of value.

    Returns False if null byte is present.

    On successful decode, checks that the amount of control codes not
    typical of text data do not exceed one third of the total characters.

    Args:
        value (bytes): bytes data
        is_truncated (bool): whether or not value has been truncated

    Returns:
        bool: True if bytes_data is likely pure text, false if likely binary
    """

    def _is_accepted_control_code_ratio(text_value):
        n = 0
        for c in text_value:
            if ord(c) == 0:
                return False
            if ord(c) < 32 and ord(c) not in _VALID_ASCII_CONTROL_CODES:
                n += 1
        return (n / len(text_value)) < _ACCEPTABLE_CONTROL_CODE_RATIO

    def _try_utf8(chunk):
        try:
            return chunk.decode("utf-8", "strict")
        except UnicodeDecodeError:
            return False

    def _utf8_decoded(value):
        if not is_truncated:
            return _try_utf8(value)
        b = value[: len(value)]
        for _ in range(4):
            if d := _try_utf8(b):
                return d
            if len(b) <= 1:
                return False
            b = b[:-1]
        return False

    if value == b"":
        return True
    if d := _utf8_decoded(value):
        return _is_accepted_control_code_ratio(d)
    return False


def random_base62(length=16):
    """Returns a safe string of sufficient length to be a nonce.

    The default is 62^16, which is 4e28. For comparison, 2^64 is 1e19 and
    2^128 is 3e38.

    Args:
        length (int): how long to make the base62 string [16]
    Returns:
        str: random base62 characters
    """
    from pykern import pkconst
    import random

    r = random.SystemRandom()
    return "".join(r.choice(pkconst.BASE62_CHARS) for x in range(length))


def unbound_localhost_tcp_port(start=10000, stop=20000):
    """Looks for AF_INET SOCK_STREAM port for which bind succeeds

    Args:
        start (int): first port [10000]
        stop (int): one greater than last port (passed to range) [20000]
    Returns:
        int: port is available or raises ValueError
    """
    import random, socket
    from pykern import pkasyncio, pkconst

    def _check_port(port):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind((pkconst.LOCALHOST_IP, int(port)))
        return port

    for p in random.sample(range(start, stop), 100):
        try:
            return _check_port(p)
        except Exception:
            pass
    raise ValueError(
        f"unable find port random sample range={start}-{stop} tries=100 ip={pkconst.LOCALHOST_IP}"
    )



"""xlsx_test case

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkcollections import PKDict
import pykern.xlsx
import decimal

PATH = "case1.xlsx"
w = pykern.xlsx.Workbook(path=PATH)
s = w.sheet(title="one")
s.table(title="types").row(
    # test all the parseable types
    int=375,
    str="a fairly long string",
    float=7.77123,
    bool=True,
    none=None,
    decimal=decimal.Decimal(3.142),
)
t = s.table(title="t1", defaults=PKDict(round_digits=2))
t.header(
    ("Left", "Middle"),
    three="Right",
    four="Very End",
)
t.row(
    Left=t.cell(
        ["*", ["-", 100, ["-", ["n"]]], 1],
        fmt="bold",
    ),
    Middle=t.cell(
        35.337,
        fmt="percent",
        link="n",
    ),
    three=t.cell(
        ["%", ["n"], 1],
        fmt="decimal",
    ),
    four=t.cell(
        ("IF", ["<=", ["n"], 2], "red", "green"),
    ),
).pkupdate(defaults=PKDict(num_fmt="currency"))
t.row(
    Left=["MAX", 999],
    Middle=["MAX", 111, 222],
    three=["MIN", 333, 444],
    four=["IF", 0, ["/", 1, 0], 99],
).pkupdate(defaults=PKDict(num_fmt="currency"))
t.row(
    Left=["AND", True, ["<=", 1, 2]],
    Middle=["OR", False, [">", 3, 4]],
    three=["NOT", ["<=", 6, 7]],
    four="no op",
)
t.footer(
    Left="L",
    Middle=None,
    three=t.cell(
        "R",
        fmt="bold",
    ),
    four=None,
)
t.footer(Left="No Totals", Middle="", three=None, four=None).pkupdate(
    defaults=PKDict(border=None),
)
w.save()



# -*- coding: utf-8 -*-
"""pytest for :mod:`pykern.pkcollections`

:copyright: Copyright (c) 2015 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
from __future__ import absolute_import, division, print_function
import pytest


def test_canonicalize():
    from pykern.pkunit import pkeq, pkexcept
    from pykern.pkcollections import PKDict, canonicalize
    import decimal

    class _i(int):
        pass

    class _o:
        pass

    pkeq(None, canonicalize(None))
    pkeq(3.3, canonicalize(decimal.Decimal(3.3)))
    pkeq(int, type(canonicalize(_i(1))))
    pkeq([1], canonicalize(set([1])))
    pkeq(["bytes", "str"], canonicalize(iter([b"bytes", "str"])))
    pkeq([9.01, True, False], canonicalize((9.01, True, False)))
    d = canonicalize({_i(-1): dict(a="b", c=frozenset([13]))})
    pkeq(PKDict({-1: PKDict(a="b", c=[13])}), d)
    pkeq(int, type(list(d.keys())[0]))
    with pkexcept("unable to canonicalize"):
        canonicalize(_o())


def test_delitem():
    from pykern.pkcollections import PKDict

    n = PKDict(a=1)
    del n["a"]
    assert not hasattr(n, "a"), "delitem should remove the item"
    with pytest.raises(KeyError):
        del n["b"]


def test_dict():
    """Validate PKDict()"""
    from pykern import pkcollections
    from pykern.pkcollections import PKDict
    from pykern.pkunit import pkok, pkeq, pkexcept

    n = PKDict()
    n.a = 1
    pkok(
        1 == n.a,
        "new attribute should work with x.y format",
    )
    pkok(
        1 == n["a"],
        'x["y"] should retrieve 1',
    )
    n.a = 2
    pkok(
        2 == n.a,
        "overwrite attr",
    )
    with pkexcept(pkcollections.PKDictNameError):
        delattr(n, "a")
    with pkexcept(pkcollections.PKDictNameError):
        setattr(n, "keys", 3)
    expect = "invalid key for PKDict"
    with pkexcept(expect):
        setattr(n, "__getattr__", 3)
    with pkexcept(pkcollections.PKDictNameError):
        delattr(n, "items")
    n = PKDict(items=1)
    pkok(list(n.items()) == [("items", 1)], "items() should be retrievable")
    n["items"] = ""
    del n["items"]
    pkok(list(n.items()) == [], "items() should be deletable")
    # specific string is generated so match exactly
    with pkexcept("'PKDict' object has no attribute 'missing_attribute'"):
        n.missing_attribute()
    with pkexcept(KeyError):
        n["missing key"]


def test_dict_copy():
    from pykern.pkcollections import PKDict
    import copy
    from pykern.pkunit import pkeq, pkne

    n = PKDict(a=1, b=PKDict(c=3))
    m = copy.copy(n)
    pkne(id(n), id(m))
    pkeq(id(n.b), id(n.b))
    m = copy.deepcopy(n)
    pkne(id(n), id(m))
    pkeq(id(n.b), id(n.b))


def test_dict_pkdel():
    from pykern.pkcollections import PKDict
    from pykern.pkunit import pkok, pkeq, pkexcept

    n = PKDict(d4=4)
    pkeq(4, n.pkdel("d4"))
    pkeq(None, n.pkdel("d4"))
    pkeq("x", n.pkdel("d4", "x"))


def test_dict_pknested_get():
    from pykern.pkcollections import PKDict
    from pykern.pkunit import pkeq, pkexcept

    n = PKDict(one=PKDict(two=1.2), simple=1)
    pkeq(1, n.pknested_get("simple"))
    pkeq(1, n.pknested_get(["simple"]))
    pkeq(1.2, n.pknested_get("one.two"))
    pkeq(1.2, n.pknested_get(("one", "two")))
    with pkexcept(TypeError):
        n.pknested_get("simple.not")
    pkeq(1, n.pkunchecked_nested_get("simple"))
    pkeq(1.2, n.pkunchecked_nested_get("one.two"))
    pkeq(None, n.pkunchecked_nested_get("one.two.three"))
    pkeq(None, n.pkunchecked_nested_get("simple.not"))
    pkeq(36, n.pkunchecked_nested_get("simple.not", 36))
    n = PKDict(one=[10, PKDict(last="done"), 14])
    pkeq("done", n.pknested_get("one.1.last"))


def test_dict_pknested_set():
    from pykern.pkcollections import PKDict
    from pykern.pkunit import pkeq, pkexcept

    n = PKDict(one=PKDict())
    n.pknested_set("one.two", 2)
    n.pknested_set("other", 3)
    pkeq(2, n.pknested_get("one.two"))
    pkeq(3, n.pknested_get(["other"]))


def test_dict_pksetdefault():
    from pykern.pkcollections import PKDict
    from pykern.pkunit import pkok, pkeq, pkexcept

    n = PKDict()
    pkeq(13, n.pksetdefault("d1", lambda: 13).d1)
    pkeq(13, n.pksetdefault("d1", "already set").d1)
    pkeq(n, n.pksetdefault("d1", "already set", "d2", 2, "d3", 3, "d4", 4))
    pkeq(2, n.d2)
    pkeq(3, n.d3)
    pkeq(4, n.d4)
    for i in "already set", 2, 3, 4:
        with pkexcept(KeyError):
            n[i]
    with pkexcept(AssertionError):
        n.pksetdefault("a", "b", "c")
    n = PKDict()
    pkeq(13, n.pksetdefault(d1=13).d1)
    pkeq(13, n.pksetdefault(d1="already set").d1)
    pkeq(n, n.pksetdefault(d1="already set", d2=2, d3=3, d4=4))
    pkeq(2, n.d2)
    pkeq(3, n.d3)
    pkeq(4, n.d4)
    for i in "already set", 2, 3, 4:
        with pkexcept(KeyError):
            n[i]
    # ok to call empty kwargs
    n.pksetdefault()


def test_dict_pksetdefault1():
    from pykern.pkcollections import PKDict
    from pykern.pkunit import pkok, pkeq, pkexcept

    n = PKDict()
    pkeq(13, n.pksetdefault1("d1", lambda: 13))
    pkeq(13, n.pksetdefault1("d1", "already set"))
    pkeq("value", n.pksetdefault1(key1="value"))
    pkeq("other", n.pksetdefault1(key2=lambda: "other"))
    for a, k, r in (
        (("d2", 2, "d3", 3), dict(), "too many args"),
        ((), dict(d2=2, d3=3), "too many kwargs"),
        ((), dict(), "no args"),
        (("d3",), dict(), "one arg"),
    ):
        with pkexcept(AssertionError, r):
            n.pksetdefault1(*a, **k)


def test_dict_pkupdate():
    from pykern.pkcollections import PKDict
    from pykern.pkunit import pkeq

    d = PKDict()
    pkeq(id(d), id(d.pkupdate(a=1)))
    pkeq(1, d.a)


def test_pkmerge():
    from pykern.pkunit import pkeq
    from pykern.pkcollections import PKDict

    s = PKDict(one=PKDict(two=None, three=PKDict(four=[4])), five=99)
    s.pkmerge(PKDict(five=5, one=PKDict(two=[1, 2], three=PKDict(four2=4.2))))
    pkeq(PKDict(one=PKDict(two=[1, 2], three=PKDict(four=[4], four2=4.2)), five=5), s)


def test_subclass():
    from pykern import pkcollections
    from pykern.pkunit import pkeq, pkok
    import copy

    class S(pkcollections.PKDict):
        def __init__(self, some_arg):
            self._some_arg = some_arg

        def __copy__(self):
            return self.__class__(self._some_arg)

    e = S(["anything"])
    a = copy.deepcopy(e)
    pkeq(e._some_arg, a._some_arg)
    pkok(id(e._some_arg) != id(a._some_arg), "some args is not copied")


def test_unchecked_del():
    from pykern.pkunit import pkeq
    from pykern import pkcollections

    n = {"a": 1, "b": 2, "c": 3}
    pkcollections.unchecked_del(n, "a")
    pkeq({"b": 2, "c": 3}, n)
    pkcollections.unchecked_del(n, "a", "b", "c")
    pkeq({}, n)



"""Wrapper for setuptools.setup to simplify creating of `setup.py` files.

Python `setup.py` files should be short for well-structured projects.
`b_setup.setup` assumes there are directories such as `tests`, `docs`,
`bin`, etc. PyKern Projects use `py.test` so the appropriate `Test`
class is provided by this module.

Example:

    A sample ``setup.py`` script::

        setup(
            name='pyexample',
            description='Some Example app',
            author='Example, Inc.',
            author_email='somebody@example.com',
            url='http://example.com',
        )

Assumptions:

    - GUI and console scripts are
      found automatically by special suffixes ``_gui.py`` and
      ``_console.py``. See ``setup`` documentation for an example.

    - Under git control. Even if you are building an app for the first
      time, you should create the repo first. Does not assume anything
      about the remote (i.e. need not be a GitHub repo).

:copyright: Copyright (c) 2015 Radiasoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# DO NOT import __future__. setuptools breaks with unicode in PY2:
# http://bugs.python.org/setuptools/issue152
# Get errors about package_data not containing wildcards, name not found, etc.

# Import only builtin/standard packages so avoid dependency issues
import copy
import datetime
import distutils.cmd
import distutils.log
import glob
import locale
import os
import os.path
import packaging.version
import re
import setuptools
import setuptools.command.sdist
import setuptools.command.test
import subprocess
import sys

#: The subdirectory in the top-level Python where to put resources
PACKAGE_DATA = "package_data"

#: Where scripts live, you probably don't want this
SCRIPTS_DIR = "scripts"

_VERSION_RE = r"(\d{8}\.\d+)"

_cfg = None


class SDist(setuptools.command.sdist.sdist, object):
    """Fix up a few things before running sdist"""

    def check_readme(self, *args, **kwargs):
        """Avoid README error message. We assert differntly.

        Currently only supports ``README.txt`` and ``README``,
        but we may have ``README.md``.
        """
        pass


def install_requires():
    """Parse requirements.txt.

    Returns:
        dict: parsed requirements.txt
    """
    res = []
    # TODO(robnagler) deprecate this for literal install_requires
    with open("requirements.txt", "r") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            assert not line.endswith("\\"), "does not support continuation lines"
            res.append(line)
    return res


def setup(**kwargs):
    """Parses `README.*` and `requirements.txt`, sets some defaults, then
    calls `setuptools.setup`.

    Scripts are found by looking for files in the top level package directory
    which end with ``_console.py`` or ``_gui.py``. These files must have a
    function called ``main``.

    Example:
        The file ``pykern_console.py`` might contain::

            def main():
                return 2 + 2

        This would create a program called command line program ``pykern`` which
        would call ``main()`` when invoked.

    Args:
        kwargs: see `setuptools.setup`
    """

    def _assert_package_versions():
        """Raise assertion if another module has installed incompatible versions

        Currently no incompatible versions that need to be asserted. This
        commit has an example of how to code this:

        https://git.radiasoft.org/pykern/commit/28c0b69034dd96785964fd7049cc5d33a5c0b9b5
        """
        pass

    name = kwargs["name"]
    if name != "pykern":
        _assert_package_versions()
    assert (
        type(name) == str
    ), "name must be a str; remove __future__ import unicode_literals in setup.py"
    flags = kwargs["pksetup"] if "pksetup" in kwargs else {}
    if "install_requires" not in kwargs:
        kwargs["install_requires"] = install_requires()
    # If the incoming is unicode, this works in Python3
    # https://bugs.python.org/issue13943
    del kwargs["name"]
    base = {
        "classifiers": [],
        "cmdclass": {
            "sdist": SDist,
        },
        "entry_points": _entry_points(name),
        # These both need to be set
        "name": name,
        "packages": _packages(name),
        "pksetup": flags,
    }
    base = _state(base, kwargs)
    _merge_kwargs(base, kwargs)
    _extras_require(base)
    op = setuptools.setup
    if base["pksetup"].get("numpy_distutils", False):
        import numpy.distutils.core

        op = numpy.distutils.core.setup
    del base["pksetup"]
    op(**base)


def _check_output(*args, **kwargs):
    """Run `subprocess.checkout_output` and convert to str

    Args:
        args (list): pass to subprocess.check_output
    Returns:
        str: Output
    """
    try:
        res = subprocess.check_output(*args, **kwargs)
        if isinstance(res, bytes):
            res = res.decode(locale.getpreferredencoding())
            return res
    except subprocess.CalledProcessError as e:
        if hasattr(e, "output") and len(e.output):
            sys.stderr.write(e.output)
        raise


def _entry_points(pkg_name):
    """Find all *_{console,gui}.py files and define them

    Args:
        pkg_name (str): name of the package (directory)

    Returns:
        dict: Mapping of script names to module:methods
    """
    res = {}
    for s in ["console", "gui"]:
        tag = "_" + s
        for p in glob.glob(os.path.join(pkg_name, "*" + tag + ".py")):
            m = re.search(
                r"^([a-z]\w+)" + tag, os.path.basename(p), flags=re.IGNORECASE
            )
            if m:
                ep = res.setdefault(s + "_scripts", [])
                # TODO(robnagler): assert that 'def main()' exists in python module
                ep.append("{} = {}.{}:main".format(m.group(1), pkg_name, m.group(0)))
    return res


def _extras_require(base):
    """Add "all" to extras_require, if supplied

    Args:
        base (dict): our base params, will be updated
    """
    if not "extras_require" in base:
        return
    er = base["extras_require"]
    if not er or "all" in er:
        return
    all_deps = set()
    for key, deps in base["extras_require"].items():
        # Explicit dependencies are not in all, e.g. ':sys_platform != "win32"'
        if ":" not in key:
            all_deps.update(deps)
    if all_deps:
        er["all"] = all_deps


def _find_files(dirname):
    """Find all files checked in with git and otherwise.

    Asserts git is installed and git repo.

    Args:
        dirname (str): directory

    Returns:
        list: Files to include in package
    """
    if _git_exists():
        res = _git_ls_files(["--others", "--exclude-standard", dirname])
        res.extend(_git_ls_files([dirname]))
    else:
        res = []
        for r, _, files in os.walk(dirname):
            for f in files:
                res.append(os.path.join(r, f))
    return sorted(res)


def _git_exists():
    """Have a git repo?

    Returns:
        bool: True if .git dir exists
    """
    return os.path.isdir(".git")


def _git_ls_files(extra_args):
    """Find all the files under git control

    Will return nothing if package_data doesn't exist or no files in it.

    Args:
        extra_args (list): other args to append to command

    Returns:
        list: Files under git control.
    """
    cmd = ["git", "ls-files"]
    cmd.extend(extra_args)
    out = _check_output(cmd, stderr=subprocess.STDOUT)
    return out.splitlines()


def _merge_kwargs(base, kwargs):
    """Merge custom values into kwargs then update base with kwargs

    Args:
        base (dict): computed defaults
        kwargs (dict): passed in from setup.py
    """
    for k in "cmdclass", "entry_points":
        if not k in kwargs:
            continue
        v = kwargs[k]
        if v:
            base[k].update(v)
        del kwargs[k]
    base.update(kwargs)


def _packages(name):
    """Find all packages by looking for ``__init__.py`` files.

    Mostly borrowed from https://bitbucket.org/django/django/src/tip/setup.py

    Args:
        name (str): name of the package (directory)

    Returns:
        list: packages names
    """

    def _fullsplit(path, result=None):
        """
        Split a pathname into components (the opposite of os.path.join) in a
        platform-neutral way.

        """
        if result is None:
            result = []
        head, tail = os.path.split(path)
        if head == "":
            return [tail] + result
        if head == path:
            return result
        return _fullsplit(head, [tail] + result)

    res = []
    for (
        dirpath,
        _,
        filenames,
    ) in os.walk(name):
        if "__init__.py" in filenames:
            res.append(str(".".join(_fullsplit(dirpath))))
    return res


def _read(filename):
    """Open and read filename

    Args:
        filename (str): what to read

    Returns:
        str: contents of filename
    """
    with open(filename, "r") as f:
        return f.read()


def _readme():
    """Find the README.*. Prefer README.rst

    Returns:
        str: Name of README
    """
    for which in "README.rst", "README.md", "README.txt":
        if os.path.exists(which):
            return which
    raise ValueError("You need to create a README.rst")


def _remove(path):
    """Remove path without throwing an exception"""
    try:
        os.remove(path)
    except OSError:
        pass


def _state(base, kwargs):
    """Gets version and package_data. Writes MANIFEST.in.

    Args:
        base (dict): our base params

    Returns:
        dict: base updated
    """
    state = {}
    sha = "\n"
    if not "version" in kwargs:
        state["version"], s = _version(base)
        if s:
            sha = "\n\ngit-commit={}\n".format(s)
    manifest = """# OVERWRITTEN by pykern.pksetup every "python setup.py"
include LICENSE
"""
    if os.path.exists("requirements.txt"):
        manifest += "include requirements.txt\n"
    readme = _readme()
    state["long_description"] = _read(readme).rstrip() + sha
    manifest += "include {}\n".format(readme)
    dirs = ["docs", "tests"]
    if "extra_directories" in base["pksetup"]:
        dirs.extend(base["pksetup"]["extra_directories"])
    for which in (PACKAGE_DATA, SCRIPTS_DIR):
        is_pd = which == PACKAGE_DATA
        d = os.path.join(base["name"], which) if is_pd else which
        f = _find_files(d)
        if f:
            if is_pd:
                state[which] = {base["name"]: f}
                state["include_package_data"] = True
            else:
                state[which] = f
            dirs.append(d)
    manifest += "".join(["recursive-include {} *\n".format(d) for d in dirs])
    _write("MANIFEST.in", manifest)
    base.update(state)
    return base


def _version(base):
    """Get a chronological version from git or PKG-INFO

    Args:
        base (dict): state

    Returns:
        str: Chronological version "yyyymmdd.hhmmss"
        str: git sha if available
    """
    from pykern import pkconfig

    global _cfg

    if not _cfg:
        _cfg = pkconfig.init(no_version=(False, bool, "use utcnow as version"))
    if _cfg.no_version:
        return _version_from_datetime(), None
    v1 = _version_from_pkg_info(base)
    v2, sha = _version_from_git(base)
    if v1:
        if v2:
            return (v1, None) if float(v1) > float(v2) else (v2, sha)
        return v1, None
    if v2:
        return v2, sha
    raise ValueError("Must have a git repo or an source distribution")


def _version_float(value):
    m = re.search(_VERSION_RE, value)
    assert m, "version={} syntax incorrect must match {}".format(value, _VERSION_RE)
    return m.group(1)[: -len(m.group(2))] if m.group(2) else m.group(1)


def _version_from_datetime(value=None):
    # Avoid 'UserWarning: Normalizing' by setuptools
    return str(
        packaging.version.Version(
            (value or datetime.datetime.utcnow()).strftime("%Y%m%d.%H%M%S"),
        ),
    )


def _version_from_git(base):
    """Chronological version string for most recent commit or time of newer file.

    Finds the commit date of the most recent branch. Uses ``git
    ls-files`` to find files under git control which are modified or
    to be deleted, in which case we assume this is a developer, and we
    should just use the current time for the version. It will be newer
    than any committed version, which is all we care about for upgrades.

    Args:
        base (dict): state

    Returns:
        str: Chronological version "yyyymmdd.hhmmss"
    """
    if not _git_exists():
        return None, None
    # Under development?
    sha = None
    if len(_git_ls_files(["--modified", "--deleted"])):
        vt = None
    else:
        branch = _check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"]).rstrip()
        vt = _check_output(["git", "log", "-1", "--format=%ct", branch]).rstrip()
        vt = datetime.datetime.fromtimestamp(float(vt))
        sha = _check_output(["git", "rev-parse", "HEAD"]).rstrip()
    return _version_from_datetime(vt), sha


def _version_from_pkg_info(base):
    """Extra existing version from PKG-INFO if there

    Args:
        base (dict): state

    Returns:
        str: Chronological version "yyyymmdd.hhmmss"
    """
    try:
        d = _read(base["name"] + ".egg-info/PKG-INFO")
        m = re.search(r"Version:\s*{}\s".format(_VERSION_RE), d)
        if m:
            return m.group(1)
    except IOError:
        pass


def _write(filename, content):
    """Writes a file"""
    with open(filename, "w") as f:
        f.write(content)



"""Python and YAML configuration file parser.

FConf reads Python and YAML files and produces a single, merged
PKDict. FConf is not a replacement for `pykern.pkconfig`. Rather it is
for complex configuration input files to programs that often require
programmatic generation. FConf was written for
`RSConf <https://git.radiasoft.org/rsconf>`_.

The Basic YAML configuration look like this::

    fconf_macros:
      make_uri(host):
          "https://${host}"

    some:
       - item
    uri: make_uri('sirepo.com')

Macros in the `fconf_macros` are simple replacements.  They cannot have
loops or other conditions. These are reserved for Python files.

A macro use must be a full text element. It cannot be embedded in
text, e.g.  ``curl make_uri('sirepo.com')``, that's because macros can
return data structures. The syntax is pure Python. Unlike YAML strings,
the arguments have to be quoted, because they will be evaluated by the Python
interpreter.

More complex macros can be programmed in a Python file::

    import math

    def hypotenuse(self, a, b):
        return math.sqrt(a ** 2, b ** 2)

Using in YAML looks the same::

    long_side: hypotenuse(3, 4)

Note that the arguments were not quoted here, because they are integers.

The typical usage is::

    fconf.parse_all(some_dir)

where `some_dir` is a `py.path` that contains python and yaml files to parse.
You can also call `Parser` directly with the a list of files to parse.

As noted, the arguments to a macro are Python. Actually, the entire
value is a Python expression, and the only requirement is that it
begin with a Python word, which can be any Python function. For example::

    count_down:
      - reversed(range(4))

Produces the following Python Structure::

    {'count_down': [3, 2, 1, 0]}

This allows for a lot of possibilities within the YAML alone.

The Python and YAML macros can call each other. Inside a Python file,
just use self to call something::

    def count_up(self, top):
        return reversed(self.count_down(top))

Note that the above count_down example used a special feature of
result merging. When a Macro returns a list (generator, tuple, etc.)
in a list context, it will merge with the list. Strings are used in
place. For example::

    mixed:
      - range(3)
      - make_uri('sirepo.com')

Produces::

    {'mixed': [0, 1, 2, 'https://sirepo.com']}

The same is true of dicts. For example::

    fconf_macros:
      base(num):
        v${num}.radia.run:
          description: VM ${num}

    all_hosts:
      base(3):
      base(5):

Produces::

    {
        "all_hosts": {
            "v3.radia.run": {
                "description": "VM 3",
            },
            "v5.radia.run": {
                "description": "VM 5",
            },
        },
    }

There are some other useful features. You can refer to any previous
declared value using variable expansion. For example::

    a:
      b:
        c: 3
    d: ${a}
    e: ${a.b.c}


Produces::

    {
        'a': {'b': {'c': 3}},
        'd': {'b': {'c': 3}},
        'e': 3,
    }

This allows for flexible constants. Those global strings can be used
in YAML Macros.

YAML files are evaluated before they are merged. However, they are all
merged before the next file is evaluated. This allows a main
"constants" file, for example, to direct the flow of the subsequent files.

Workarounds
-----------

Most strings do not need to be quoted in YAML, but with the extra syntax of FConf,
there are some special cases, for example, this fails::

  a: 1
  b: [ ${a} ]


It gets the error ``ruamel.yaml.parser.ParserError: while parsing a
flow sequence in "<unicode string> did not find expected ',' or
']'``. To work around, simply put in quotes::

  a: 1
  b: [ "${a}" ]

Another case is inline Python with braces like this::

  a: c({"d": "e"})

Just quote with single quotes::

  a: 'c({"d": "e"})'

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html

"""

from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp
import contextlib
import copy
import inspect
import pykern.pkrunpy
import re
import pykern.pkyaml
import pykern.pkcollections
import pykern.pkinspect
import pykern.pkio


#: matches word(word1, ...). Words have to begin with a letters
_TEXT_MACRO_DEF = re.compile(
    r"^([a-z]\w*)\(((?:\s*[a-z]\w*\s*)?(?:,\s*[a-z]\w*\s*)*)\)$", flags=re.IGNORECASE
)

_MACRO_CALL = re.compile(r"^([a-z]\w*)\((.*)\)$", flags=re.IGNORECASE + re.DOTALL)

_ARG_SEP = re.compile(r"\s*,\s*|\s+")

#: Matches result value and macro argument expansions. FVAR stands for Fconf Variable
_FVAR = re.compile(r"\$\{(\S+)\}")

#: Use an exact pattern so we can know that the result could be a data structure
_FVAR_EXACT = re.compile(f"^{_FVAR.pattern}$")

_RESERVED_PREFIX = "fconf_"

_SELF = _RESERVED_PREFIX + "self"

_NO_PARAM = object()

_BUILTINS_EXT = "builtins"


class Parser(PKDict):
    def __init__(self, files, base_vars=None):
        self.pkupdate(
            files=PKDict(),
            macros=PKDict(),
        )
        self._add_file(pykern.pkio.py_path(f"fconf.{_BUILTINS_EXT}"))
        for f in files:
            try:
                self._add_file(f)
            except Exception as e:
                pykern.pkinspect.append_exception_reason(e, f"fconf.Parser.file={f}")
                raise
        if "yml" not in self.files:
            raise ValueError("must supply at least one '.yml' file")
        e = _Evaluator(
            global_fvars=PKDict() if base_vars is None else copy.deepcopy(base_vars),
            parser=self,
        )
        for f in self.files.yml:
            e.start(source=f)
        self.result = e.global_fvars

    def _add_file(self, path):
        e = path.ext[1:].lower()
        f = _File(
            content=getattr(self, f"_ext_{e}")(path),
            ext=e,
            path=path,
        )
        if e in ("py", _BUILTINS_EXT):
            self._add_macros(f)
        elif e == "yml":
            self._add_text_macros(f)
        else:
            raise ValueError(f"unhandled file ext={e}")
        self.files.setdefault(e, []).append(f)

    def _add_macro(self, macro):
        n = macro.name
        if n in self.macros:
            raise ValueError(f"duplicate {macro}, other {self.macros[macro.name]}")
        if n.startswith(_RESERVED_PREFIX) and macro.source.ext != _BUILTINS_EXT:
            raise ValueError(f"macro={macro} may not begin with {_RESERVED_PREFIX}")
        self.macros[macro.name] = macro

    def _add_macros(self, source):
        m = source.content
        if not m:
            return
        for n, f in m.items():
            self._add_macro(
                _Macro(
                    func=f,
                    name=n,
                    # TODO(robnagler) assert varnames don't begin with fconf_
                    params=tuple(f.__code__.co_varnames),
                    source=source,
                ),
            )

    def _add_text_macros(self, source):
        m = source.content.pkdel("fconf_macros")
        if not m:
            return
        for n, c in m.items():
            m = _TEXT_MACRO_DEF.search(n)
            if not m:
                raise ValueError(f"invalid macro name={n} {source}")
            n = m.group(1)
            # TODO: duplicate arg check
            self._add_macro(
                _YAMLMacro(
                    content=c,
                    name=n,
                    # TODO(robnagler): check for dups
                    # TODO(robnagler): assert varnames don't begin with fconf_
                    params=tuple((x for x in _ARG_SEP.split(m.group(2)) if x)),
                    source=source,
                ),
            )

    def _ext_builtins(self, path):
        return self._functions(_Builtins)

    def _ext_py(self, path):
        return self._functions(
            pykern.pkrunpy.run_path_as_module(path), co_filename=str(path)
        )

    def _ext_yml(self, path):
        return pykern.pkyaml.load_file(path)

    def _functions(self, obj, co_filename=None):
        import types

        m = PKDict()
        for n, o in inspect.getmembers(obj):
            # Ensure that macro comes from the file, not an import
            if inspect.isfunction(o) and (
                not co_filename or co_filename == o.__code__.co_filename
            ):
                m[n] = o
        return m


def parse_all(path, base_vars=None):
    """Parse all the Python and YAML files in `directory`

    Files are read in sorted order with all Python files first and
    YAML files next. YAML file evaluation happens in that same order.

    Args:
        path (py.path): directory that ``*.py`` and ``*.yml`` files
        base_vars (PKDict): initial variable state. May be hierarchical. [None]
    Returns:
        PKDict: evaluated and merged files plus base_vars
    """
    return Parser(
        pykern.pkio.sorted_glob(path.join("*.py"))
        + pykern.pkio.sorted_glob(path.join("*.yml")),
        base_vars=None,
    ).result


class _Builtins:
    @staticmethod
    def fconf_var(namespace, name):
        return namespace._evaluator.fconf_var(name)


class _Evaluator(PKDict):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # By NOT passing __builtins__=PKDict(), builtins are implicitly added.
        # Python builtins are useful
        self.global_ns = PKDict()
        self.local_ns = _Namespace(self)
        self.local_fvars = PKDict()

    def fconf_var(self, name):
        try:
            return self.global_fvars.pknested_get(name)
        except KeyError:
            # Do not cascade the exception.
            pass
        raise KeyError(f"unknown macro param or fconf_var={name}")

    def start(self, source, **kwargs):
        p = self.local_fvars
        i = "local_fvars" in kwargs
        if i:
            self.local_fvars = kwargs["local_fvars"]
            base = None
        else:
            if not (source.content is None or isinstance(source.content, PKDict)):
                raise ValueError(f"{source} must be PKDict or empty")
            base = self.global_fvars
            self.xpath = _XPath()
        try:
            with self._xpath(kwargs.get("xpath_element")):
                return self._do(source.content, base)
        except Exception as e:
            if not i:
                pykern.pkinspect.append_exception_reason(
                    e,
                    f"fconf._Evaluator.source={source}\n{self.xpath.stack_as_str()}",
                )
            raise
        finally:
            self.local_fvars = p

    def _dict(self, new, base):
        if not isinstance(base, PKDict):
            if base is not None:
                raise ValueError(f"mismatched types new={new} base={base}")
            base = PKDict()
        for k, v in new.items():
            # TODO: pass v to expr
            with self._xpath(k):
                x = self._expr(k)
                with self._xpath(None if x == k else x):
                    if isinstance(x, PKDict):
                        # TODO better error msgs
                        assert v is None, f"key={k} must not have a value"
                        base.pkmerge(x, make_copy=False)
                    elif isinstance(x, list):
                        raise ValueError(
                            f"mismatched types expanding macro={k} to value={x}"
                            + " into dict={res}",
                        )
                    else:
                        assert x is not None, f"expanded macro={k} to None"
                        base[x] = self._do(v, base.get(x))
        return base

    def _do(self, new, base):
        if isinstance(new, PKDict):
            return self._dict(new, base)
        if isinstance(new, list):
            return self._list(new, base)
        return self._expr(new)

    def _expr(self, value):
        def _fvar_op(match, native=False, use_repr=False):
            n = match.group(1)
            if n in self.local_fvars:
                res = self.local_fvars[n]
            else:
                res = self.fconf_var(n)
            if use_repr:
                return repr(res)
            return res if native else str(res)

        def _fvar_sub(value, use_repr=False):
            with self._xpath(value):
                m = None if use_repr else _FVAR_EXACT.search(value)
                # This check prevents stringification of data types on exact
                # matches, which is important for non-string fvars
                if m:
                    res = _fvar_op(m, native=True)
                    if not isinstance(res, str):
                        # already canonicalized
                        return True, res
                else:
                    res = _FVAR.sub(
                        lambda x: _fvar_op(x, use_repr=use_repr),
                        value,
                    )
                return False, res

        if not isinstance(value, str):
            # already canonicalized
            return value
        k, v = _fvar_sub(value)
        if k:
            return v
        with self._xpath(v):
            m = _MACRO_CALL.search(v)
        if not m:
            return v
        with self._xpath(v):
            a = _fvar_sub(m.group(2))[1] if len(m.group(2)) > 0 else ""
            s = f"{_SELF}," if m.group(1) in self.parser.macros else ""
            v = f"{m.group(1)}({s}{a})"
            with self._xpath(v):
                return pykern.pkcollections.canonicalize(
                    eval(v, self.global_ns, self.local_ns),
                )

    def _list(self, new, base):
        if not isinstance(base, list):
            if base is not None:
                raise ValueError(f"mismatched types new={new} base={base}")
            base = []
        res = []
        for i, e in enumerate(new):
            # lists don't have base values
            with self._xpath(f"[{i}]"):
                x = self._do(e, None)
            if isinstance(x, list):
                res.extend(x)
            else:
                res.append(x)
        return res + base

    @contextlib.contextmanager
    def _xpath(self, element=None):
        """Track element path for errors

        If there is an exception, xpath will hold the place of the exception
        """
        if element is not None:
            self.xpath.push(element)
            pkdc("{}", element)
            yield
            self.xpath.pop()
        else:
            yield


class _File(PKDict):
    def __str__(self):
        return "source=" + pykern.pkio.py_path().bestrelpath(self.path)


class _Macro(PKDict):
    def call(self, namespace, args, kwargs):
        return self.func(namespace, *args, **kwargs)

    def _kind(self):
        return "pymacro"

    def __str__(self):
        return f"{self._kind}={self.name} {self.source}"


class _Namespace:
    def __init__(self, evaluator):
        self._evaluator = evaluator

    def __func(self, name, exc):
        def _canonicalize(value):
            """Allow functions at the top level

            There are cases when it is useful to pass functions even to YAML macros.
            """
            if inspect.isfunction(value):
                return value
            return pykern.pkcollections.canonicalize(value)

        def _wrapper(*args, **kwargs):
            a = list(args)
            if a and a[0] is self:
                a.pop(0)
            return m.call(
                self,
                [_canonicalize(x) for x in a],
                PKDict({k: _canonicalize(v) for k, v in kwargs.items()}),
            )

        m = self._evaluator.parser.macros.get(name)
        if not m:
            raise exc(f"macro name={name}")

        return _wrapper

    def __getattr__(self, name):
        return self.__func(name, AttributeError)

    def __getitem__(self, name):
        if name == _SELF:
            return self
        return self.__func(name, KeyError)


class _XPath:
    def __init__(self):
        self.stack = []

    def push(self, key):
        f = inspect.currentframe().f_back.f_back.f_back
        self.stack.append(
            PKDict(key=key, func=f.f_code.co_name, line=f.f_lineno),
        )

    def pop(self):
        self.stack.pop()

    def __str__(self):
        return "/".join([k.key for k in self.stack])

    def stack_as_str(self):
        res = "Evaluator stack:\n"
        for i in self.stack:
            res += self._str(i)
        return res

    def _str(self, element):
        return f"{element.func}:{element.line} {str(element.key):.100}\n"


class _YAMLMacro(PKDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # TEST: duplicate params, invalid params, etc.
        s = set()
        for p in self.params:
            if p in s:
                raise TypeError(f"duplicate param={p} in macro={self.name}")
            s.add(p)

    def call(self, namespace, args, kwargs):
        return namespace._evaluator.start(
            base=None,
            local_fvars=self._parse_args(args, kwargs),
            source=self,
            xpath_element=self.name + "()",
        )

    def _kind(self):
        return "macro"

    def _parse_args(self, args, kwargs):
        p = list(self.params)
        res = PKDict({k: _NO_PARAM for k in p})
        # DOC: there are no optional params in text_macros
        for a in args:
            if not p:
                raise TypeError(f"too many args={args} macro={self.name}")
            res[p.pop(0)] = a
        for k, v in kwargs.items():
            if k not in p:
                if k not in self.params:
                    raise TypeError(f"invalid kwarg={k} macro={self.name}")
                raise TypeError(f"position arg followed by kwarg={k} macro={self.name}")
            if res[k] is not _NO_PARAM:
                raise TypeError(f"duplicate kwarg={k} macro={self.name}")
            res[k] = v
        x = [k for k, v in res.items() if v is _NO_PARAM]
        if x:
            raise TypeError(f"missing args={x} macro={self.name}")
        return res

    def __str__(self):
        return f"macro={self.name}"



"""test restartable

:copyright: Copyright (c) 2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""


def test_restarable():
    from pykern import pkio
    from pykern import pkunit
    from pykern.pkcli import test

    # ensure does not abort for max_failures (only two tests)
    test._cfg.max_failures = 3
    for p in (1, 2, 3):
        for test._cfg.restartable in (False, True):
            with pkunit.save_chdir_work() as d:
                pkunit.data_dir().join("tests").copy(d.join("tests"))
                with pkunit.pkexcept(
                    "FAILED=1" if test._cfg.restartable else "FAILED=2",
                ):
                    test.default_command(f"max_procs={p}")



"""Helper functions for to :mod:`inspect`.

:copyright: Copyright (c) 2015 RadiaSoft, Inc.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

# Avoid pykern imports so avoid dependency issues for pkconfig
from pykern.pkcollections import PKDict
import importlib
import inspect
import os
import os.path
import pkgutil
import re
import sys

#: Used to simplify paths output
_start_dir = ""
try:
    _start_dir = os.getcwd()
except Exception:
    pass


_VALID_IDENTIFIER_RE = re.compile(r"^[a-z_]\w*$", re.IGNORECASE)


class Call(PKDict):
    """Saves file:line:name of stack frame and renders as string.

    Args:
        frame_or_log (frame or LogRecord): values to extract

    Attributes:
        filename (str): full path (co_filename)
        lineno (int): line number (f_lineno)
        name (str): function name (co_name)
    """

    def __init__(self, frame_or_log):
        try:
            if hasattr(frame_or_log, "f_code"):
                super(Call, self).__init__(
                    filename=frame_or_log.f_code.co_filename,
                    lineno=frame_or_log.f_lineno,
                    name=frame_or_log.f_code.co_name,
                    # Only used by caller_module()
                    _module=sys.modules.get(frame_or_log.f_globals.get("__name__")),
                )
            else:
                super(Call, self).__init__(
                    filename=frame_or_log.pathname,
                    lineno=frame_or_log.lineno,
                    name=frame_or_log.funcName,
                    _module=None,
                )
        finally:
            if frame_or_log:
                del frame_or_log

    def __str__(self):
        try:
            filename = os.path.relpath(self.filename, _start_dir)
            if len(filename) > len(self.filename):
                # "relpath" always makes relative even when no common components.
                # Take the absolute (shorter) path
                filename = self.filename
            return "{}:{}:{}".format(filename, self.lineno, self.name)
        except Exception:
            return "<no file>:0:<no func>"


def append_exception_reason(exc, reason):
    """Augment `exc` with `reason`

    Modifies `exc` in place by adding to `exc.args` or
    `exc.reason`. Does it's best to not cause another exception during
    this process.

    Args:
        exc (BaseException): what was raised
        reason (str): our related reason

    """

    def _prefix_reason(string):
        return ("; " if len(string) > 0 else "") + reason

    if hasattr(exc, "reason") and isinstance(exc.reason, str):
        exc.reason += _prefix_reason(exc.reason)
    if hasattr(exc, "args"):
        if exc.args is None:
            exc.args = tuple()
        if isinstance(exc.args, (tuple, list)):
            if len(exc.args) == 0:
                exc.args = (reason,)
            elif isinstance(exc.args[0], str):
                x = list(exc.args)
                x[0] += _prefix_reason(x[0])
                exc.args = tuple(x)
    # Add other cases as they arise
    # Otherwise, leave exception unmodified


def caller(ignore_modules=None, exclude_first=True):
    """Which file:line:func is calling the caller of this function.

    Will not return the same module as the calling module, that is,
    will iterate until a new module is found. If `ignore_modules` is
    defined, will ignore those modules as well.

    Note: may return __main__ module.

    Will raise exception if calling from __main__

    Args:
        ignore_modules (list): other modules (objects) to exclude [None]
        exclude_first (bool): skip first module found [True]

    Returns:
        PKDict: keys: filename, lineno, name, module
    """
    frame = None
    try:
        exclude = [inspect.getmodule(caller)]
        if ignore_modules:
            exclude.extend(ignore_modules)
        exclude_orig_len = len(exclude)
        # Ugly code, because don't want to bind "frame"
        # in a call.
        frame = inspect.currentframe().f_back
        while True:
            m = inspect.getmodule(frame)
            # getmodule doesn't always work for some reason
            if not m:
                m = sys.modules[frame.f_globals["__name__"]]
            if m not in exclude:
                if len(exclude) > exclude_orig_len or not exclude_first:
                    return Call(frame)
                # Have to go back two exclusions (this module and our caller)
                exclude.append(m)
            frame = frame.f_back
        # Will raise exception if calling from __main__
    finally:
        # If an exception is thrown, the stack
        # hangs around forever. That's what the del frame
        # is for.
        if frame:
            del frame


def caller_func_name():
    """Name of function one frame back

    Useful for inter-module dispatch and errors.

    Returns:
        str: function name
    """
    return inspect.currentframe().f_back.f_back.f_code.co_name


def caller_module(exclude_first=True):
    """Which module is calling the caller of this function.

    Will not return the same module as the calling module, that is,
    will iterate until a new module is found. If exclude_first == True
    it will also exclude the first module found that is not the calling
    module.

    Note: may return __main__ module.

    Will raise exception if calling from __main__

    Args:
        exclude_first (bool): skip first module found [True]

    Returns:
        module: module which is calling module
    """
    return caller(exclude_first=exclude_first)._module


def is_caller_main():
    """Is the caller's calling module __main__?

    Returns:
        bool: True if calling module was called by __main__.
    """
    return caller_module().__name__ == "__main__"


def is_valid_identifier(string):
    """Is this a valid Python identifier?

    Args:
        string (str): what to validate
    Returns:
        bool: True if is valid python ident.
    """
    return bool(_VALID_IDENTIFIER_RE.search(string))


def module_basename(obj):
    """Parse the last part of a module name

    For example, module_basename(pkinspect) is 'pkinspect'.

    Args:
        obj (object): any python object

    Returns:
        str: base part of the module name
    """
    return module_name_split(obj).pop()


def module_name_join(names):
    """Joins names with '.'

    Args:
        names (iterable): list of strings to join

    Returns:
        str: module name
    """
    return ".".join(names)


def module_name_split(obj):
    """Splits obj's module name on '.'

    Args:
        obj (object): any python object

    Returns:
        str: base part of the module name
    """
    n = inspect.getmodule(obj).__name__
    return n.split(".")


def module_functions(func_prefix, module=None):
    """Get all module level functions starting with func_prefix

    Args:
        func_prefix (str): the prefix of function names to get
        module (object): a module to get functions from (calling module if None)

    Returns:
        PKDict: dict of function name mapped to the function object
    """
    r = PKDict()
    for n, o in inspect.getmembers(module or caller_module(exclude_first=False)):
        if n.startswith(func_prefix) and inspect.isfunction(o):
            r[n] = o
    return r


def package_module_names(name_or_module):
    """List of modules of package `name_or_module`

    Args:
        name_or_module (object): absolute name of package, e.g. pykern.pkcli, or module object
    Returns:
        list: sorted, relative module names, e.g. [ci, fmt, github, ...]
    """
    p = (
        name_or_module
        if inspect.ismodule(name_or_module)
        else importlib.import_module(name_or_module)
    )
    return sorted(
        (
            m.name
            for m in pkgutil.iter_modules([os.path.dirname(p.__file__)])
            if not m.ispkg
        ),
    )


def root_package(obj):
    """Parse the root package in which `obj` is defined.

    For example, root_package(module_basename) is 'pykern'.

    Args:
        obj (object): any python object

    Returns:
        str: root package for the object
    """
    return module_name_split(obj).pop(0)


def submodule_name(obj):
    """Remove the root package in which `obj` is defined.

    For example, root_package(module_basename) is 'pkinspect'.

    Args:
        obj (object): any python object

    Returns:
        str: submodule for the object
    """
    x = module_name_split(obj)
    x.pop(0)
    return module_name_join(x)


def this_module():
    """Module object for caller

    Returns:
        module: module object
    """
    return caller(exclude_first=False)._module



# -*- coding: utf-8 -*-
"""test pkconfig.in_dev_mode()

:copyright: Copyright (c) 2023 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""
import pytest


def test_in_dev_mode():
    """Validate channel_in()"""
    from pykern import pkconfig

    assert pkconfig.in_dev_mode(), "should be in dev mode"



"""WebSocket Quest server

:copyright: Copyright (c) 2025 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdc, pkdlog, pkdp, pkdexc
from pykern.api import util
import asyncio
import importlib
import inspect
import pykern.pkasyncio
import pykern.quest
import pykern.util
import tornado.websocket
import re


_API_NAME_RE = re.compile(rf"^{pykern.quest.API.METHOD_PREFIX}(\w+)")


class Session(pykern.quest.Attr):
    """State held on server bound to a client.

    Currently the state is not persisted when the server terminates. This may change.
    """

    ATTR_KEY = "session"

    IS_SINGLETON = True

    def handle_on_close(self):
        x = list(self.values())
        # Reversed so LIFO
        while x:
            if s := getattr(x.pop(), "session_end", None):
                s()


class Subscription(pykern.quest.Attr):
    """EXPERIMENTAL"""

    ATTR_KEY = "subscription"

    def __init__(self, server_msg):
        super().__init__(None, _server_msg=server_msg)

    def result_put(self, api_result):
        self._server_msg.subscription_result_put(api_result)


def start(api_classes, attr_classes, http_config, coros=()):
    """Start `_Server` in `pkasyncio`

    Args:
        api_classes (Iterable): `pykern.quest.API` subclasses to be dispatched
        attr_classes (Iterable): `pykern.quest.Attr` subclasses to create API instance
        http_config (PKDict): `pkasyncio.Loop.http_server` arg
        coros (Iterable): list of coroutines to be passed to `pkasyncio.Loop.run`
    """
    l = pykern.pkasyncio.Loop()
    _Server(l, api_classes, attr_classes, http_config)
    if coros:
        l.run(*coros)
    l.start()


class _Server:
    def __init__(self, loop, api_classes, attr_classes, http_config):
        def _api_class_funcs():
            a = False
            for c in api_classes:
                for r in _api_class_funcs1(c):
                    if r.name == util.AUTH_API_NAME:
                        a = True
                    yield r
            if not a:
                for r in _api_class_funcs1(
                    importlib.import_module("pykern.api.auth_api").AuthAPI
                ):
                    yield r

        def _api_class_funcs1(clazz):
            for n, o in inspect.getmembers(clazz, predicate=inspect.isfunction):
                if not (m := _API_NAME_RE.search(n)):
                    continue
                yield PKDict(
                    class_=clazz,
                    func=o,
                    func_name=n,
                    is_subscription=util.is_subscription(o),
                    name=m.group(1),
                )

        def _api_map():
            rv = PKDict()
            for a in _api_class_funcs():
                if a.name in rv:
                    raise AssertionError(
                        "duplicate api={a.name} class={a.class_.__name__}"
                    )
                # don't need to save func
                if not inspect.iscoroutinefunction(a.pkdel("func")):
                    raise AssertionError(
                        "api_func={n} is not async class={a.class_.__name__}"
                    )
                rv[a.name] = a
            return rv

        h = http_config.copy().pksetdefault(uri_map=[])
        self.loop = loop
        self.api_map = _api_map()
        self.attr_classes = attr_classes
        h.uri_map = h.uri_map[:]
        h.uri_map.append((h.api_uri, _ServerHandler, PKDict(server=self)))
        self.api_uri = h.pkdel("api_uri")
        h.log_function = self._log_end
        self._ws_id = 0
        loop.http_server(h)

    def handle_get(self, handler):
        self._log(handler, "ws-get")

    def handle_open(self, handler):
        try:
            self._ws_id += 1
            handler.pykern_api_context.ws_id = self._ws_id
            return _ServerConnection(self, handler, ws_id=self._ws_id)
        except Exception as e:
            pkdlog("exception={} stack={}", e, pkdexc())
            self._log(handler, "open-error", "exception={}", [e])
            return None

    def _log(self, handler, which, fmt="", args=None):
        def _add(key, value):
            nonlocal f, a
            if value is not None:
                f += (" " if f else "") + key + "={}"
                a.append(value)

        f = ""
        a = []
        if x := getattr(handler, "pykern_api_context", None):
            _add("error", x.pkdel("error"))
            _add("ws_id", x.get("ws_id"))
        if fmt:
            f = f + " " + fmt
            a.extend(args)
        self.loop.http_log(handler, which, f, a)

    def _log_end(self, handler, *args, **kwargs):
        if isinstance(handler, _ServerHandler):
            self._log(handler, "ws-end")
        else:
            self.loop.http_log(handler)


class _ServerConnection:

    def __init__(self, server, handler, ws_id):
        self.server = server
        self.handler = handler
        self.ws_id = ws_id
        self.pending_msgs = []
        self._destroyed = False
        self.session = Session(None)
        self.remote_peer = server.loop.remote_peer(handler.request)
        self.log("ws-open")

    def destroy(self):
        if self._destroyed:
            return
        self._destroyed = True
        x = list(self.pending_msgs)
        self.pending_msgs = []
        while x:
            # Reversed so end in opposite order of calls (LIFO)
            x.pop().destroy()
        self.handler.close()
        self.handler = None
        if s := self.session:
            self.session = None
            # Last since it calls out of this module
            s.handle_on_close()

    def handle_on_close(self):
        self.destroy()

    async def handle_on_message(self, msg):
        if self._destroyed:
            return
        m = None
        try:
            m = _ServerMsg(self)
            self.pending_msgs.append(m)
            if not await m.process(msg):
                self.destroy()
        except Exception as e:
            pkdlog("exception={} stack={}", e, pkdexc())
            self.log("msg-error", m, "unhandled exception={}", [e])
            self.destroy()
        finally:
            try:
                if not self._destroyed and m:
                    # Maybe have been already destroyed
                    self.pending_msgs.remove(m)
            except Exception as e:
                pkdlog("exception={} stack={}", e, pkdexc())
                pass

    def log(self, which, call=None, fmt="", args=None):
        if fmt:
            fmt = " " + fmt
        pkdlog(
            "{} ip={} ws={} {}" + fmt,
            which,
            self.remote_peer,
            self.ws_id,
            call,
            *(args if args else ()),
        )


class _ServerHandler(tornado.websocket.WebSocketHandler):
    def initialize(self, server):
        # Since part of a global space, need to prefix
        self.pykern_api_server = server
        self.pykern_api_context = PKDict()
        self.pykern_api_connection = None

    async def get(self, *args, **kwargs):
        try:
            self.pykern_api_server.handle_get(self)
            return await super().get(*args, **kwargs)
        except Exception as e:
            pkdlog("exception={} stack={}", e, pkdexc())

    async def on_message(self, msg):
        try:
            # WebSocketHandler only allows one on_message at a time
            pykern.pkasyncio.create_task(
                self.pykern_api_connection.handle_on_message(msg)
            )
        except Exception as e:
            pkdlog("exception={} stack={}", e, pkdexc())

    def on_close(self):
        try:
            if not (c := self.pykern_api_connection):
                return
            self.pykern_api_connection = None
            c.handle_on_close()
        except Exception as e:
            pkdlog("exception={} stack={}", e, pkdexc())

    def open(self):
        try:
            self.pykern_api_connection = self.pykern_api_server.handle_open(self)
        except Exception as e:
            pkdlog("exception={} stack={}", e, pkdexc())


class _ServerMsg:

    def __init__(self, connection):
        self._connection = connection
        self._call = None
        self._qcall = None
        self._api = None
        self._destroyed = False

    def destroy(self, unsubscribe=False):
        if self._destroyed:
            return
        self._destroyed = True
        if not (c := self._qcall):
            return
        self._qcall = None
        c.quest_end(in_error=not unsubscribe)

    async def process(self, msg):
        try:
            self._call, e = util.msg_unpack(msg, "server")
            if e:
                self._log("unpack-error", "error={}", [e])
                return False
            if r := self._parse():
                pass
            elif self._call.msg_kind.is_unsubscribe():
                self._unsubscribe(self._call.call_id)
                return True
            elif self._call.msg_kind.is_subscribe():
                r = await self._do_call(Subscription(self))
                if r is not None and not isinstance(r, Exception):
                    r = util.APICallError(
                        f"return type={type(r)} from subscription api must be None"
                    )
            else:
                r = await self._do_call(None)
            if self._destroyed:
                return True
            self._reply(r)
            if isinstance(r, util.APIProtocolError):
                self._log("protocol-error", "exception={}", [e])
                return False
            self._log("end")
            return True
        except Exception as e:
            pkdlog("exception={} {} stack={}", e, self, pkdexc())
            self._log("process-error", "exception={}", [e])
            self._reply(e)
            return False
        finally:
            self.destroy()

    def subscription_result_put(self, api_result):
        if isinstance(api_result, Exception):
            raise util.APICallError(
                "api_result type={type(api_result)} may not be an exception"
            )
        if api_result is None:
            raise util.APICallError("api_result may not be None")
        self._reply(api_result)
        if self._destroyed:
            raise util.APIDisconnected()

    async def _do_call(self, sub):
        try:
            # Let quest.start see the exception
            with self._quest_start(sub) as c:
                try:
                    self._qcall = c
                    return await getattr(c, self._api.func_name)(self._call.api_args)
                finally:
                    self._qcall = None
        except Exception as e:
            pkdlog("exception={} {} stack={}", self, e, pkdexc())
            if not isinstance(e, pykern.util.APIError):
                e = util.APICallError(f"unhandled_exception={e}")
            return e

    def _log(self, which, *args):
        return self._connection.log(which, self, *args)

    def _parse(self):
        def _args():
            if self._call.get("api_args") is None:
                return util.APIProtocolError("missing msg field api_args")
            return None

        def _name():
            if not (n := self._call.get("api_name")):
                return util.APIProtocolError("missing msg field api_name")
            if a := self._connection.server.api_map.get(n):
                self._api = a
                return None
            return util.APINotFound(n)

        def _kind():
            k = self._call.msg_kind
            if k.is_unsubscribe():
                return None
            if r := _name():
                return r
            if k.is_subscribe():
                if not self._api.is_subscription:
                    return util.APIKindError(
                        f"cannot subscribe non-subscription api={self._call.api_name}"
                    )
            elif k.is_call():
                if self._api.is_subscription:
                    return util.APIKindError(
                        f"call_api on subscription api={self._call.api_name}"
                    )
            else:
                raise AssertionError(f"invalid {k} returned from msg_unpack")
            return _args()

        self._log(self._call.msg_kind.name.lower())
        return _kind()

    def _quest_start(self, sub=None):
        a = list(self._connection.server.attr_classes)
        a.append(self._connection.session)
        if sub:
            a.append(sub)
        return pykern.quest.start(self._api.class_, a)

    def _reply(self, call_rv):
        try:
            if call_rv == None:
                if self._call.msg_kind.is_subscribe():
                    r = PKDict(msg_kind=util.MsgKind.UNSUBSCRIBE)
                else:
                    r = PKDict(api_result=None, api_error="missing reply")
            elif not isinstance(call_rv, Exception):
                r = PKDict(api_result=call_rv, api_error=None)
            elif isinstance(call_rv, pykern.util.APIError):
                r = PKDict(
                    api_result=None,
                    api_error=f"{call_rv.__class__.__name__}: {call_rv}",
                )
            else:
                r = PKDict(api_result=None, api_error=f"unhandled_exception={call_rv}")
            r.pksetdefault(msg_kind=util.MsgKind.REPLY)
            r.call_id = self._call.call_id
            self._connection.handler.write_message(util.msg_pack(r), binary=True)
            self._log("reply")
        except Exception as e:
            pkdlog("exception={} {} stack={}", e, self, pkdexc())
            self._log("reply-error")
            self.destroy()

    def __str__(self):
        def _destroyed():
            return " DESTROYED" if self._destroyed else ""

        def _info(c):
            if not c:
                return
            k = c.get("msg_kind")
            i = c.get("call_id")
            if not (k and i):
                return
            rv = f"{k.name.lower()}#{i}"
            if a := c.get("api_name"):
                rv += " " + a
            return f"<{rv}{_destroyed()}>"

        return _info(self._call) or f"<{self.__class__.__name__}{_destroyed()}>"

    def _unsubscribe(self, call_id):
        for m in self._connection.pending_msgs:
            if m._call and m._call.get("call_id") == call_id and m._api:
                if m._api.is_subscription:
                    m.destroy(unsubscribe=True)



"""Wrapper on asyncio and tornado

:copyright: Copyright (c) 2022 RadiaSoft LLC.  All Rights Reserved.
:license: http://www.apache.org/licenses/LICENSE-2.0.html
"""

from pykern import pkconfig
from pykern import pkconst
from pykern.pkcollections import PKDict
from pykern.pkdebug import pkdlog, pkdp
import asyncio
import inspect
import re
import tornado.gen
import tornado.httpserver
import tornado.ioloop
import tornado.web

_cfg = None

_background_tasks = set()


class Loop:
    def __init__(self):
        _init()
        self._coroutines = []
        self._http_server = False

    def http_server(self, http_cfg):
        """Instantiate a tornado web server

        Under the covers Tornado uses the asyncio event loop so asyncio methods
        can be mixed with Tornado methods.

        Using asyncio methods, e.g. `asyncio.run`, is preferred over
        Tornado methods, e.g.  `tornado.ioloop.IOLoop.current` to
        reduce dependency on Tornado. Using this module should allow
        the code to be portable to other http server frameworks.

        ``http_config.uri_map`` maps URI expressions to classes, which
        is passed directly to `tornado.web.Application`.

        Args:
            http_cfg (PKDict): quest_start, uri_map, debug, tcp_ip, tcp_port,

        """

        async def _do():
            # TODO(e-carlin): pull in the one in job_supervisor.py
            p = http_cfg.get("tcp_port", _cfg.server_port)
            i = http_cfg.get("tcp_ip", _cfg.server_ip)
            tornado.httpserver.HTTPServer(
                tornado.web.Application(
                    http_cfg.uri_map,
                    debug=http_cfg.get("debug", _cfg.debug),
                    log_function=http_cfg.get("log_function", self.http_log),
                ),
                xheaders=True,
            ).listen(p, i)
            pkdlog("name={} ip={} port={}", http_cfg.get("name"), i, p)
            await asyncio.Event().wait()

        if self._http_server:
            raise AssertionError("http_server may only be called once")
        self._http_server = True
        self.run(_do())

    def http_log(self, handler, which="end", fmt="", args=None):
        r = handler.request
        f = "{} ip={} uri={}"
        a = [which, self.remote_peer(r), r.uri]
        if fmt:
            f += " " + fmt
            a += args
        if which == "start":
            f += " proto={} {} ref={} ua={}"
            a += [
                r.method,
                r.version,
                r.headers.get("Referer") or "",
                r.headers.get("User-Agent") or "",
            ]
        elif which == "end":
            f += " status={} ms={:.2f}"
            a += [
                handler.get_status(),
                r.request_time() * 1000.0,
            ]
        pkdlog(f, *a)

    def remote_peer(self, request):
        # https://github.com/tornadoweb/tornado/issues/2967#issuecomment-757370594
        # implementation may change; Code in tornado.httputil check connection.
        if c := request.connection:
            # socket is not set on stream for websockets.
            if getattr(c, "stream", None) and (s := getattr(c.stream, "socket", None)):
                return "{}:{}".format(*s.getpeername())
        i = request.headers.get("proxy-for", request.remote_ip)
        return f"{i}:0"

    def run(self, *coros):
        for c in coros:
            if not inspect.iscoroutine(c):
                raise AssertionError(f"must be a coroutine arg={c} coros={coros}")
        try:
            asyncio.get_running_loop()
        except RuntimeError:
            self._coroutines.extend(coros)
            return
        raise AssertionError("cannot call after event loop has started")

    def start(self):
        async def _do():
            await asyncio.gather(*self._coroutines)

        if not self._coroutines:
            raise AssertionError("no coroutines registered; must have at least one")
        asyncio.run(_do(), debug=_cfg.debug)


@pkconfig.parse_none
def cfg_ip(value):
    if value is None:
        return "0.0.0.0" if pkconfig.in_dev_mode() else pkconst.LOCALHOST_IP
    return value


def cfg_port(value):
    v = int(value)
    l = 3000
    u = 32767
    if not l <= v <= u:
        pkconfig.raise_error(f"value must be from {l} to {u}")
    return v


def create_task(coro):
    """Create a task

    Keeps a global reference to the task so to avoid the garbage
    collector running before the task is run.
    https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
    """
    t = asyncio.create_task(coro)
    _background_tasks.add(t)
    t.add_done_callback(_background_tasks.discard)
    return t


async def sleep(secs):
    await asyncio.sleep(secs)


def _init():
    global _cfg
    if _cfg:
        return
    _cfg = pkconfig.init(
        debug=(False, bool, "enable debugging for asyncio"),
        server_ip=(
            None,
            cfg_ip,
            "ip to listen on",
        ),
        server_port=("9001", cfg_port, "port to listen on"),
        verify_tls=(
            not pkconfig.channel_in("dev"),
            bool,
            "validate TLS certificates on requests; for self-signed set to False",
        ),
    )
